"""Implementation of the state design pattern for transfers"""
from aiofiles import os as asyncos
import asyncio
from enum import Enum
import inspect
import logging
from types import MethodType
from typing import Optional, Protocol, TYPE_CHECKING

if TYPE_CHECKING:
    from .model import Transfer


logger = logging.getLogger(__name__)


def _with_state_lock(func):

    async def wrapper(obj: 'TransferState', *args, **kwargs):
        async with obj.transfer._state_lock:
            result = await func(*args, **kwargs)
        return result

    return wrapper


async def _remove_local_file(transfer: 'Transfer'):
    if not transfer.is_download():
        return

    if transfer.local_path:
        logger.info("removing file from filesystem : %s", transfer.local_path)
        try:
            if await asyncos.path.exists(transfer.local_path):
                await asyncos.remove(transfer.local_path)
        except OSError:
            logger.warning("failed to remove file during abort : %s", transfer.local_path)

        transfer.local_path = None


class TransferStateListener(Protocol):

    async def on_transfer_state_changed(
            self, transfer: 'Transfer', old: 'TransferState.State', new: 'TransferState.State'):
        ...


class TransferState:
    """Represents a transfer state and its possible transitions

    Each transition method will return a boolean to indicate whether the
    transition was done.
    """

    class State(Enum):
        UNSET = -1
        VIRGIN = 0
        QUEUED = 1
        INITIALIZING = 3
        INCOMPLETE = 4
        DOWNLOADING = 5
        UPLOADING = 6
        COMPLETE = 7
        FAILED = 8
        ABORTED = 9
        PAUSED = 10

    UNSET = State.UNSET
    VIRGIN = State.VIRGIN
    QUEUED = State.QUEUED
    INITIALIZING = State.INITIALIZING
    INCOMPLETE = State.INCOMPLETE
    DOWNLOADING = State.DOWNLOADING
    UPLOADING = State.UPLOADING
    COMPLETE = State.COMPLETE
    FAILED = State.FAILED
    ABORTED = State.ABORTED
    PAUSED = State.PAUSED

    VALUE = UNSET

    def __init__(self, transfer: 'Transfer'):
        self.transfer: 'Transfer' = transfer
        self._wrap_lock()

    @classmethod
    def init_from_state(cls, state: State, transfer: 'Transfer'):
        for subcls in cls.__subclasses__():
            if subcls.VALUE == state:
                return subcls(transfer)

        raise Exception(f"no state class for state : {state}")

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._wrap_lock()

    async def fail(self, reason: Optional[str] = None) -> bool:  # pragma: no cover
        logger.warning(
            "attempted to make undefined state transition from %s to %s", self.VALUE.name, self.FAILED.name)
        return False

    async def abort(self, reason: Optional[str] = None) -> bool:  # pragma: no cover
        logger.warning(
            "attempted to make undefined state transition from %s to %s", self.VALUE.name, self.ABORTED.name)
        return False

    async def queue(self, remotely: bool = False) -> bool:  # pragma: no cover
        logger.warning(
            "attempted to make undefined state transition from %s to %s", self.VALUE.name, self.QUEUED.name)
        return False

    async def initialize(self) -> bool:  # pragma: no cover
        logger.warning(
            "attempted to make undefined state transition from %s to %s", self.VALUE.name, self.INITIALIZING.name)
        return False

    async def complete(self) -> bool:  # pragma: no cover
        logger.warning(
            "attempted to make undefined state transition from %s to %s", self.VALUE.name, self.COMPLETE.name)
        return False

    async def incomplete(self) -> bool:  # pragma: no cover
        logger.warning(
            "attempted to make undefined state transition from %s to %s", self.VALUE.name, self.INCOMPLETE.name)
        return False

    async def start_transferring(self) -> bool:  # pragma: no cover
        new_state = self.UPLOADING if self.transfer.is_upload() else self.DOWNLOADING
        logger.warning(
            "attempted to make undefined state transition from %s to %s", self.VALUE.name, new_state.name)
        return False

    async def pause(self) -> bool:  # pragma: no cover
        logger.warning(
            "attempted to make undefined state transition from %s to %s", self.VALUE.name, self.PAUSED.name)
        return False

    def _wrap_lock(self):
        """Decorates all methods with a lock"""
        for name, method in inspect.getmembers(self, predicate=inspect.ismethod):
            if not name.startswith('_'):
                setattr(self, name, MethodType(_with_state_lock(method), self))

    async def _cancel_transfer_tasks(self):
        await asyncio.gather(*self.transfer.cancel_tasks(), return_exceptions=True)

    async def _stop_transfer(self):
        await self._cancel_transfer_tasks()
        self.transfer.set_complete_time()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.transfer!r})"


class VirginState(TransferState):
    """State representing a newly added transfer"""
    VALUE = TransferState.VIRGIN

    async def queue(self, remotely: bool = False) -> bool:
        self.transfer.remotely_queued = remotely
        await self.transfer.transition(QueuedState(self.transfer))
        return True

    async def pause(self) -> bool:
        await self.transfer.transition(PausedState(self.transfer))
        return True


class QueuedState(TransferState):
    VALUE = TransferState.QUEUED

    async def initialize(self) -> bool:
        await self.transfer.transition(InitializingState(self.transfer))
        return True

    async def fail(self, reason: Optional[str] = None) -> bool:
        self.transfer.fail_reason = reason
        await self.transfer.transition(FailedState(self.transfer))
        return True

    async def abort(self, reason: Optional[str] = None) -> bool:
        await self._cancel_transfer_tasks()
        await _remove_local_file(self.transfer)
        self.transfer.abort_reason = reason
        await self.transfer.transition(AbortedState(self.transfer))
        return True

    async def pause(self) -> bool:
        await self._cancel_transfer_tasks()
        await self.transfer.transition(PausedState(self.transfer))
        return True


class InitializingState(TransferState):
    """Initializing state:

    - Uploads: This indicates we are attempting to establish a connection to the
    peer to start uploading a file
    - Downloads: The download will quickly go into this state when the transfer
    ticket has been received over a file connection
    """
    VALUE = TransferState.INITIALIZING

    async def abort(self, reason: Optional[str] = None) -> bool:
        await self._cancel_transfer_tasks()
        await _remove_local_file(self.transfer)
        self.transfer.abort_reason = reason
        await self.transfer.transition(AbortedState(self.transfer))
        return True

    async def pause(self) -> bool:
        await self._cancel_transfer_tasks()
        await self.transfer.transition(PausedState(self.transfer))
        return True

    async def queue(self, remotely: bool = False):
        self.transfer.remotely_queued = remotely
        await self.transfer.transition(QueuedState(self.transfer))
        return True

    async def fail(self, reason: Optional[str] = None) -> bool:
        self.transfer.fail_reason = reason
        await self.transfer.transition(FailedState(self.transfer))
        return True

    async def start_transferring(self) -> bool:
        self.transfer.set_start_time()
        self.transfer.reset_queue_vars()
        if self.transfer.is_upload():
            await self.transfer.transition(UploadingState(self.transfer))
        else:
            await self.transfer.transition(DownloadingState(self.transfer))
        return True


class DownloadingState(TransferState):
    VALUE = TransferState.DOWNLOADING

    async def fail(self, reason: Optional[str] = None) -> bool:
        self.transfer.fail_reason = reason
        self.transfer.set_complete_time()
        await self.transfer.transition(FailedState(self.transfer))
        return True

    async def complete(self) -> bool:
        self.transfer.set_complete_time()
        await self.transfer.transition(CompleteState(self.transfer))
        return True

    async def abort(self, reason: Optional[str] = None) -> bool:
        await self._stop_transfer()
        await _remove_local_file(self.transfer)
        self.transfer.abort_reason = reason
        await self.transfer.transition(AbortedState(self.transfer))
        return True

    async def pause(self) -> bool:
        await self._stop_transfer()
        await self.transfer.transition(PausedState(self.transfer))
        return True

    async def incomplete(self) -> bool:
        self.transfer.set_complete_time()
        await self.transfer.transition(IncompleteState(self.transfer))
        return True


class UploadingState(TransferState):
    VALUE = TransferState.UPLOADING

    async def fail(self, reason: Optional[str] = None) -> bool:
        self.transfer.fail_reason = reason
        self.transfer.set_complete_time()
        await self.transfer.transition(FailedState(self.transfer))
        return True

    async def complete(self) -> bool:
        self.transfer.set_complete_time()
        await self.transfer.transition(CompleteState(self.transfer))
        return True

    async def abort(self, reason: Optional[str] = None) -> bool:
        await self._stop_transfer()
        # Don't remove file
        self.transfer.abort_reason = reason
        await self.transfer.transition(AbortedState(self.transfer))
        return True

    async def pause(self) -> bool:
        await self._stop_transfer()
        await self.transfer.transition(PausedState(self.transfer))
        return True


class CompleteState(TransferState):
    VALUE = TransferState.COMPLETE

    async def queue(self, remotely: bool = False) -> bool:
        self.transfer.reset_time_vars()

        if self.transfer.is_download():
            self.transfer.reset_progress_vars()
            self.transfer.reset_local_vars()

        self.transfer.remotely_queued = remotely

        await self.transfer.transition(QueuedState(self.transfer))
        return True


class IncompleteState(TransferState):
    """State only used for downloads. The transfer should enter this state if an
    error occured during transferring but there was no explicit error from the
    other peer. In this state it should be possible to retry the transfer.
    """
    VALUE = TransferState.INCOMPLETE

    async def fail(self, reason: Optional[str] = None) -> bool:
        self.transfer.fail_reason = reason
        await self.transfer.transition(FailedState(self.transfer))
        return True

    async def queue(self, remotely: bool = False) -> bool:
        self.transfer.reset_time_vars()
        self.transfer.remotely_queued = remotely
        await self.transfer.transition(QueuedState(self.transfer))
        return True

    async def initialize(self) -> bool:
        self.transfer.reset_time_vars()
        await self.transfer.transition(InitializingState(self.transfer))
        return True

    async def abort(self, reason: Optional[str] = None) -> bool:
        await self._cancel_transfer_tasks()
        await _remove_local_file(self.transfer)
        self.transfer.abort_reason = reason
        await self.transfer.transition(AbortedState(self.transfer))
        return True

    async def pause(self) -> bool:
        await self._cancel_transfer_tasks()
        await self.transfer.transition(PausedState(self.transfer))
        return True


class FailedState(TransferState):
    VALUE = TransferState.FAILED

    async def queue(self, remotely: bool = False) -> bool:
        self.transfer.reset_time_vars()
        self.transfer.remotely_queued = remotely
        self.transfer.fail_reason = None
        self.transfer.abort_reason = None
        await self.transfer.transition(QueuedState(self.transfer))
        return True


class PausedState(TransferState):
    VALUE = TransferState.PAUSED

    async def queue(self, remotely: bool = False) -> bool:
        self.transfer.reset_time_vars()
        self.transfer.remotely_queued = remotely
        await self.transfer.transition(QueuedState(self.transfer))
        return True

    async def abort(self, reason: Optional[str] = None) -> bool:
        await self._cancel_transfer_tasks()
        await _remove_local_file(self.transfer)
        self.transfer.abort_reason = reason
        await self.transfer.transition(AbortedState(self.transfer))
        return True

    async def fail(self, reason: Optional[str] = None) -> bool:
        self.transfer.fail_reason = reason
        await self.transfer.transition(FailedState(self.transfer))
        return True


class AbortedState(TransferState):
    VALUE = TransferState.ABORTED

    async def queue(self, remotely: bool = False) -> bool:
        self.transfer.reset_time_vars()

        if self.transfer.is_download():
            self.transfer.reset_progress_vars()
            self.transfer.reset_local_vars()

        self.transfer.fail_reason = None
        self.transfer.abort_reason = None
        self.transfer.remotely_queued = remotely
        await self.transfer.transition(QueuedState(self.transfer))
        return True



"""Definition of all SoulSeek protocol messages.

This file contains 2 types of messages:

* Server

    Request : used from client to server. A client will only use the
        ``serialize`` method of these messages
    Response : used from server to client. A client will only use the
        ``deserialize`` method of these messages

* Peer

    Request : peer messages only consist of request type messages. The client
        should use the ``deserialize`` method of these messages upon receiving
        data from another peer and the ``serialize`` method when sending to
        another peer
"""
from dataclasses import dataclass, field
import logging
from typing import ClassVar, Optional

from ..exceptions import UnknownMessageError
from .primitives import (
    boolean,
    bytearr,
    uint8,
    uint16,
    uint32,
    uint64,
    string,
    array,
    ipaddr,
    FileData,
    DirectoryData,
    MessageDataclass,
    UserStats,
    PotentialParent,
    SimilarUser,
    Recommendation,
    RoomTicker,
)


logger = logging.getLogger(__name__)


class ServerMessage:
    """Class for identifying server messages"""

    @classmethod
    def deserialize_request(cls, message: bytes):
        _, msg_id = uint32.deserialize(4, message)

        for msg_class in cls.__subclasses__():
            request_cls = getattr(msg_class, 'Request', None)
            if request_cls and request_cls.MESSAGE_ID == msg_id:
                return request_cls.deserialize(0, message)

        raise UnknownMessageError(msg_id, message, "Unknown server request message")

    @classmethod
    def deserialize_response(cls, message):
        _, msg_id = uint32.deserialize(4, message)

        for msg_class in cls.__subclasses__():
            response_cls = getattr(msg_class, 'Response', None)
            if response_cls and response_cls.MESSAGE_ID == msg_id:
                return response_cls.deserialize(0, message)

        raise UnknownMessageError(msg_id, message, "Unknown server response message")


class PeerInitializationMessage:
    """Class for identifying peer initialization messages"""

    @classmethod
    def deserialize_request(cls, message: bytes):
        _, msg_id = uint8.deserialize(4, message)

        for msg_class in cls.__subclasses__():
            request_cls = getattr(msg_class, 'Request', None)
            if request_cls and request_cls.MESSAGE_ID == msg_id:
                return request_cls.deserialize(0, message)

        raise UnknownMessageError(msg_id, message, "Unknown peer initialization message")


class PeerMessage:
    """Class for identifying peer messages"""

    @classmethod
    def deserialize_request(cls, message: bytes):
        _, msg_id = uint32.deserialize(4, message)

        for msg_class in cls.__subclasses__():
            request_cls = getattr(msg_class, 'Request', None)
            if request_cls and request_cls.MESSAGE_ID == msg_id:
                return request_cls.deserialize(0, message)

        raise UnknownMessageError(msg_id, message, "Unknown peer message")


class DistributedMessage:
    """Class for identifying distributed messages"""

    @classmethod
    def deserialize_request(cls, message: bytes):
        _, msg_id = uint8.deserialize(4, message)

        for msg_class in cls.__subclasses__():
            request_cls = getattr(msg_class, 'Request', None)
            if request_cls and request_cls.MESSAGE_ID == msg_id:
                return request_cls.deserialize(0, message)

        raise UnknownMessageError(msg_id, message, "Unknown distributed message")


class Login(ServerMessage):
    """Login into the server, this should be the first message sent to the
    server upon connecting

    * The ``md5hash`` parameter in the request is the MD5 hash of the
      concatenated ``username`` and ``password``
    * The ``md5hash`` parameter in the response is the MD5 hash of the
      ``password``

    .. note::

        Older client versions (at least 149 or below) would not send the
        ``md5hash`` and ``minor_version``

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x01)
        username: str = field(metadata={'type': string})
        password: str = field(metadata={'type': string})
        client_version: int = field(metadata={'type': uint32})
        md5hash: str = field(metadata={'type': string})
        minor_version: int = field(metadata={'type': uint32})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x01)
        success: bool = field(metadata={'type': boolean})
        greeting: Optional[str] = field(default=None, metadata={'type': string, 'if_true': 'success'})
        ip: Optional[str] = field(default=None, metadata={'type': ipaddr, 'if_true': 'success'})
        md5hash: Optional[str] = field(default=None, metadata={'type': string, 'if_true': 'success'})
        privileged: Optional[bool] = field(default=None, metadata={'type': boolean, 'if_true': 'success'})
        reason: Optional[str] = field(default=None, metadata={'type': string, 'if_false': 'success'})


class SetListenPort(ServerMessage):
    """Advertise our listening ports to the server

    Obfuscated port: this part seems to be optional, either it can be omitted
    completely or both values set to ``0``

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x02)
        port: int = field(metadata={'type': uint32})
        obfuscated_port_amount: Optional[int] = field(default=None, metadata={'type': uint32, 'optional': True})
        obfuscated_port: Optional[int] = field(default=None, metadata={'type': uint32, 'optional': True})


class GetPeerAddress(ServerMessage):
    """Retrieve the IP address/port of a peer. Obfuscated port: this part is
    optional, either it can be omitted completely or both values set to ``0`` to
    indicate there is no obfuscated port

    If the peer does not exist or is not logged on the server will respond with
    IP address set to ``0.0.0.0``, port set to ``0``

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x03)
        username: str = field(metadata={'type': string})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x03)
        username: str = field(metadata={'type': string})
        ip: str = field(metadata={'type': ipaddr})
        port: int = field(metadata={'type': uint32})
        obfuscated_port_amount: Optional[int] = field(default=None, metadata={'type': uint32, 'optional': True})
        obfuscated_port: Optional[int] = field(default=None, metadata={'type': uint16, 'optional': True})


class AddUser(ServerMessage):
    """When a user is added with this message the server will automatically send
    user status updates using the :ref:`GetUserStatus` message.

    When a user sends a message to multiple users using the
    :ref:`PrivateChatMessageUsers` message then this message will only be
    received if the sender was added first using this message.

    To remove a user use the :ref:`RemoveUser` message. Keep in mind that you
    will still receive status updates in case you are joined in the same room
    with the user.

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x05)
        username: str = field(metadata={'type': string})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x05)
        username: str = field(metadata={'type': string})
        exists: bool = field(metadata={'type': boolean})
        status: Optional[int] = field(default=None, metadata={'type': uint32, 'if_true': 'exists'})
        user_stats: Optional[UserStats] = field(default=None, metadata={'type': UserStats, 'if_true': 'exists'})
        country_code: Optional[str] = field(
            default=None,
            metadata={
                'type': string,
                'if_true': 'exists',
                'optional': True
            })


class RemoveUser(ServerMessage):
    """Remove the tracking of user status which was previously added with the
    :ref:`AddUser` message.

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x06)
        username: str = field(metadata={'type': string})


class GetUserStatus(ServerMessage):
    """Get the user status. The server will automatically send updates for users
    that we have added with :ref:`AddUser` or which we share a room with.

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x07)
        username: str = field(metadata={'type': string})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x07)
        username: str = field(metadata={'type': string})
        status: int = field(metadata={'type': uint32})
        privileged: bool = field(metadata={'type': boolean})


class IgnoreUser(ServerMessage):
    """Sent when we want to ignore a user. Received when another user ignores us

    :status: DEPRECATED, DEFUNCT
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x0B)
        username: str = field(metadata={'type': string})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x0B)
        username: str = field(metadata={'type': string})


class UnignoreUser(ServerMessage):
    """Sent when we want to unignore a user. Received when another user
    unignores us

    :status: DEPRECATED, DEFUNCT
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x0C)
        username: str = field(metadata={'type': string})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x0C)
        username: str = field(metadata={'type': string})


class RoomChatMessage(ServerMessage):
    """Used when sending a message to a room or receiving a message from someone
    (including self) who sent a message to a room

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x0D)
        room: str = field(metadata={'type': string})
        message: str = field(metadata={'type': string})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x0D)
        room: str = field(metadata={'type': string})
        username: str = field(metadata={'type': string})
        message: str = field(metadata={'type': string})


class JoinRoom(ServerMessage):
    """Used when we want to join a chat room. If the chat room does not exist
    it will be created. Upon successfully joining the room the server will send
    the response message

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x0E)
        room: str = field(metadata={'type': string})
        is_private: bool = field(default=False, metadata={'type': uint32, 'optional': True})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x0E)
        room: str = field(metadata={'type': string})
        users: list[str] = field(metadata={'type': array, 'subtype': string})
        users_status: list[int] = field(metadata={'type': array, 'subtype': uint32})
        users_stats: list[UserStats] = field(metadata={'type': array, 'subtype': UserStats})
        users_slots_free: list[int] = field(metadata={'type': array, 'subtype': uint32})
        users_countries: list[str] = field(metadata={'type': array, 'subtype': string})
        owner: Optional[str] = field(default=None, metadata={'type': string, 'optional': True})
        operators: Optional[list[str]] = field(
            default=None,
            metadata={
                'type': array,
                'subtype': string,
                'optional': True
            })


class LeaveRoom(ServerMessage):
    """Used when we want to leave a chat room. The receive message is
    confirmation that we left the room

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x0F)
        room: str = field(metadata={'type': string})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x0F)
        room: str = field(metadata={'type': string})


class UserJoinedRoom(ServerMessage):
    """Received when a user joined a room

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x10)
        room: str = field(metadata={'type': string})
        username: str = field(metadata={'type': string})
        status: int = field(metadata={'type': uint32})
        user_stats: UserStats = field(metadata={'type': UserStats})
        slots_free: int = field(metadata={'type': uint32})
        country_code: str = field(metadata={'type': string})


class UserLeftRoom(ServerMessage):
    """Received when a user left a room

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x11)
        room: str = field(metadata={'type': string})
        username: str = field(metadata={'type': string})


class ConnectToPeer(ServerMessage):
    """Received when a peer attempted to connect to us but failed and thus is
    asking us to attempt to connect to them through the server. Likewise when
    we cannot connect to peer we should send this message to indicate to the
    other peer that he should try connecting to us

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x12)
        ticket: int = field(metadata={'type': uint32})
        username: str = field(metadata={'type': string})
        typ: str = field(metadata={'type': string})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x12)
        username: str = field(metadata={'type': string})
        typ: str = field(metadata={'type': string})
        ip: str = field(metadata={'type': ipaddr})
        port: int = field(metadata={'type': uint32})
        ticket: int = field(metadata={'type': uint32})
        privileged: bool = field(metadata={'type': boolean})
        obfuscated_port_amount: Optional[int] = field(default=None, metadata={'type': uint32, 'optional': True})
        obfuscated_port: Optional[int] = field(default=None, metadata={'type': uint32, 'optional': True})


class PrivateChatMessage(ServerMessage):
    """Send or receive a private message. The ``chat_id`` should be used in the
    :ref:`PrivateChatMessageAck` message to acknowledge the message has been
    received. If the acknowledgement is not sent the server will repeat the
    message on the next logon.

    The ``is_direct`` boolean indicates whether it is the first attempt to send
    the message, if the server retries then this parameter will be false

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x16)
        username: str = field(metadata={'type': string})
        message: str = field(metadata={'type': string})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x16)
        chat_id: int = field(metadata={'type': uint32})
        timestamp: int = field(metadata={'type': uint32})
        username: str = field(metadata={'type': string})
        message: str = field(metadata={'type': string})
        is_direct: Optional[bool] = field(default=False, metadata={'type': boolean, 'optional': True})

        @property
        def is_admin(self) -> Optional[bool]:
            """Only kept to keep backward compatibility. Use ``is_direct``

            This property does not actually represent whether the message was sent
            by an admin. Instead it represents whether the message was sent directly
            or was queued on the server before being sent (example, when user was
            offline and came back online)
            """
            return self.is_direct

        @is_admin.setter
        def is_admin(self, value: bool):
            self.is_direct = value


class PrivateChatMessageAck(ServerMessage):
    """Acknowledge we have received a private message after receiving a
    :ref:`PrivateChatMessage`

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x17)
        chat_id: int = field(metadata={'type': uint32})


class FileSearchRoom(ServerMessage):
    """Deprecated message for searching a room

    :Status: DEPRECATED, DEFUNCT
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x19)
        ticket: int = field(metadata={'type': uint32})
        room_id: int = field(metadata={'type': uint32})
        query: str = field(metadata={'type': string})


class FileSearch(ServerMessage):
    """This message is received when another user performed a :ref:`RoomSearch`
    or :ref:`UserSearch` request and we are part of the room or we are the user
    the user would like to search in

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x1A)
        ticket: int = field(metadata={'type': uint32})
        query: str = field(metadata={'type': string})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x1A)
        username: str = field(metadata={'type': string})
        ticket: int = field(metadata={'type': uint32})
        query: str = field(metadata={'type': string})


class SetStatus(ServerMessage):
    """Used to update the online status. Possible values for ``status``:
    online = 2, away = 1, offline = 0

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x1C)
        status: int = field(metadata={'type': uint32})


class Ping(ServerMessage):
    """Send a ping to the server to let it know we are still alive (every 5
    minutes)

    Older server versions would respond to this message with the response message

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x20)

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x20)


class SendConnectTicket(ServerMessage):
    """Deprecated predecessor to :ref:`ConnectToPeer`. A peer would send this
    message to the server when wanting to create a connection to another peer,
    the server would then pass to this to the targeted peer

    The value of the ``ticket`` parameter would be used in the :ref:`PeerInit`
    message.

    :status: DEPRECATED, DEFUNCT
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x21)
        username: str = field(metadata={'type': string})
        ticket: int = field(metadata={'type': uint32})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x21)
        username: str = field(metadata={'type': string})
        ticket: int = field(metadata={'type': uint32})


class SendDownloadSpeed(ServerMessage):
    """Sent by old clients after download has completed. :ref:`SendUploadSpeed`
    should be used instead. The ``speed`` value should be in bytes per second

    :status: DEPRECATED, DEFUNCT
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x22)
        username: str = field(metadata={'type': string})
        speed: int = field(metadata={'type': uint32})


class SharedFoldersFiles(ServerMessage):
    """Let the server know the amount of files and directories we are sharing.
    These would be returned in several messages, for example the
    :ref:`GetUserStats` and :ref:`AddUser` messages

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x23)
        shared_folder_count: int = field(metadata={'type': uint32})
        shared_file_count: int = field(metadata={'type': uint32})


class GetUserStats(ServerMessage):
    """Request a user's transfer statistics. This message will be received
    automatically for users with which we share a room

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x24)
        username: str = field(metadata={'type': string})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x24)
        username: str = field(metadata={'type': string})
        user_stats: UserStats = field(metadata={'type': UserStats})


class Kicked(ServerMessage):
    """You were kicked from the server. This message is currently only known to
    be sent when the user was logged into at another location

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x29)


class UserSearch(ServerMessage):
    """Search for a file on a specific user, the user will receive this query
    in the form of a :ref:`FileSearch` message

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x2A)
        username: str = field(metadata={'type': string})
        ticket: int = field(metadata={'type': uint32})
        query: str = field(metadata={'type': string})


class DeprecatedGetItemRecommendations(ServerMessage):
    """Similar to :ref:`GetItemRecommendations` except that no score is returned

    :status: DEPRECATED, DEFUNCT
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x32)
        item: str = field(metadata={'type': string})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x32)
        item: str = field(metadata={'type': string})
        recommendations: list[str] = field(metadata={'type': array, 'subtype': string})


class AddInterest(ServerMessage):
    """Adds an interest. This is used when requesting recommendations
    (eg.: :ref:`GetRecommendations`, ...)

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x33)
        interest: str = field(metadata={'type': string})


class RemoveInterest(ServerMessage):
    """Removes an interest previously added with :ref:`AddInterest` message

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x34)
        interest: str = field(metadata={'type': string})


class GetRecommendations(ServerMessage):
    """Request the server to send a list of recommendations and
    unrecommendations. A maximum of 100 each will be returned. The score can be
    negative.

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x36)

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x36)
        recommendations: list[Recommendation] = field(
            metadata={'type': array, 'subtype': Recommendation}
        )
        unrecommendations: list[Recommendation] = field(
            metadata={'type': array, 'subtype': Recommendation}
        )


class GetInterests(ServerMessage):
    """Request the server the list of interests it currently has stored for us.
    This was sent by older clients during logon, presumably to sync the
    interests on the client and the server. Deprecated as the client should just
    advertise all interests after logon using the :ref:`AddInterest` and
    :ref:`AddHatedInterest` messages

    Not known whether the server still responds to this command

    :status: DEPRECATED, DEFUNCT
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x37)

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x37)
        interests: list[str] = field(metadata={'type': array, 'subtype': string})


class GetGlobalRecommendations(ServerMessage):
    """Get the global list of recommendations. This does not take into account
    interests or hated interests that were previously added and is just a
    ranking of interests that other users have set

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x38)

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x38)
        recommendations: list[Recommendation] = field(
            metadata={'type': array, 'subtype': Recommendation}
        )
        unrecommendations: list[Recommendation] = field(
            metadata={'type': array, 'subtype': Recommendation}
        )


class GetUserInterests(ServerMessage):
    """Get the interests and hated interests of a particular user

    :status: DEPRECATED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x39)
        username: str = field(metadata={'type': string})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x39)
        username: str = field(metadata={'type': string})
        interests: list[str] = field(metadata={'type': array, 'subtype': string})
        hated_interests: list[str] = field(metadata={'type': array, 'subtype': string})


class ExecuteCommand(ServerMessage):
    """Send a command to the server.

    The command type has only ever been seen as having value ``admin``, the ``arguments`` array
    contains the subcommand and arguments. Example when banning a user:

    * ``command_type`` : ``admin``
    * ``arguments``

      * 0 : ``ban``
      * 1 : ``some user``
      * 2 : probably some extra args, perhaps time limit in case of ban, ... (optional)

    :status: DEPRECATED, DEFUNCT
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x3A)
        command_type: str = field(metadata={'type': string})
        arguments: list[str] = field(metadata={'type': array, 'subtype': string})


class RoomList(ServerMessage):
    """Request or receive the list of rooms. This message will be initially sent
    after logging on but can also be manually requested afterwards. The initial
    message after logon will only return a limited number of public rooms (only
    the rooms with 5 or more users).

    Parameter ``rooms_private`` excludes private rooms of which we are owner

    Parameter ``rooms_private_owned_user_count`` / ``rooms_private_user_count``
    should be the amount of users who have joined the private room, not the
    amount of members

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x40)

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x40)
        rooms: list[str] = field(metadata={'type': array, 'subtype': string})
        rooms_user_count: list[int] = field(metadata={'type': array, 'subtype': uint32})
        rooms_private_owned: list[str] = field(metadata={'type': array, 'subtype': string})
        rooms_private_owned_user_count: list[int] = field(metadata={'type': array, 'subtype': uint32})
        rooms_private: list[str] = field(metadata={'type': array, 'subtype': string})
        rooms_private_user_count: list[int] = field(metadata={'type': array, 'subtype': uint32})
        rooms_private_operated: list[str] = field(metadata={'type': array, 'subtype': string})


class ExactFileSearch(ServerMessage):
    """Used by older clients but doesn't return anything. The ``pathname`` is
    optional but is still required to be sent.

    For the message sending: The first 4 parameters are verified, the meaning
    of the final 5 bytes is unknown

    For the message receiving: message is never seen and is based on other
    documentation (PySlsk)

    :status: DEPRECATED, DEFUNCT
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x41)
        ticket: int = field(metadata={'type': uint32})
        filename: str = field(metadata={'type': string})
        pathname: str = field(metadata={'type': string})
        filesize: int = field(metadata={'type': uint64})
        checksum: int = field(metadata={'type': uint32})
        unknown: int = field(metadata={'type': uint8})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x41)
        username: str = field(metadata={'type': string})
        ticket: int = field(metadata={'type': uint32})
        filename: str = field(metadata={'type': string})
        pathname: str = field(metadata={'type': string})
        filesize: int = field(metadata={'type': uint64})
        checksum: int = field(metadata={'type': uint32})


class AdminMessage(ServerMessage):
    """Sent by the admin when the server is going down for example

    :status: UNKNOWN
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x42)
        message: str = field(metadata={'type': string})


class GetUserList(ServerMessage):
    """Gets a list of all users on the server

    :status: DEPRECATED, DEFUNCT
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x43)

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x43)
        users: list[str] = field(metadata={'type': array, 'subtype': string})
        users_status: list[int] = field(metadata={'type': array, 'subtype': uint32})
        users_stats: list[UserStats] = field(metadata={'type': array, 'subtype': UserStats})
        users_slots_free: list[int] = field(metadata={'type': array, 'subtype': uint32})
        users_countries: list[str] = field(metadata={'type': array, 'subtype': string})


class TunneledMessage(ServerMessage):
    """Tunnel a message through the server to a user

    :status: DEPRECATED, DEFUNCT
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x44)
        username: str = field(metadata={'type': string})
        ticket: int = field(metadata={'type': uint32})
        code: int = field(metadata={'type': uint32})
        message: str = field(metadata={'type': string})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x44)
        username: str = field(metadata={'type': string})
        ticket: int = field(metadata={'type': uint32})
        code: int = field(metadata={'type': uint32})
        ip: str = field(metadata={'type': ipaddr})
        port: int = field(metadata={'type': uint32})
        message: str = field(metadata={'type': string})


class PrivilegedUsers(ServerMessage):
    """List of users with privileges sent after login

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x45)
        users: list[str] = field(metadata={'type': array, 'subtype': string})


class ToggleParentSearch(ServerMessage):
    """Indicates whether we want to receive :ref:`PotentialParents` messages
    from the server. A message should be sent to disable if we have found a
    parent

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x47)
        enable: bool = field(metadata={'type': boolean})


class ParentIP(ServerMessage):
    """IP address of the parent. Not sent by newer clients

    :status: DEPRECATED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x49)
        ip: str = field(metadata={'type': ipaddr})


class Unknown80(ServerMessage):
    """Unknown message used by old client versions. The client would establish
    2 connections to the server: to one it would send the :ref:`Login` message,
    to the other this message would be sent. This second connection seemed to
    be related to the distributed network as the client would automatically
    disconnect after :ref:`DistributedAliveInterval` had been reached. It would
    seem like the server would be the client's parent in this case.

    After an interval determined by :ref:`DistributedDistributeInterval` the
    client would send another message over this connection which is described
    in :ref:`DistributedInit`. There's many unknowns in this message and even
    the types are unknown as many values were just 0, only known value is the
    last value which is the second listening port these clients used

    :status: DEPRECATED, DEFUNCT
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x50)


class ParentMinSpeed(ServerMessage):
    """Used for calculating the maximum amount of children we can have in the
    distributed network. If our average upload speed is below this value then
    we should accept no children. The average upload speed should be determined
    by the upload speed returned by :ref:`GetUserStats` (with our own username)

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x53)
        speed: int = field(metadata={'type': uint32})


class ParentSpeedRatio(ServerMessage):
    """Used for calculating the maximum amount of children we can have in the
    distributed network

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x54)
        ratio: int = field(metadata={'type': uint32})


class ParentInactivityTimeout(ServerMessage):
    """Timeout for the distributed parent

    :status: DEPRECATED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x56)
        timeout: int = field(metadata={'type': uint32})


class SearchInactivityTimeout(ServerMessage):

    """Presumably indicates after how much time search responses should no
    longer be accepted

    :status: DEPRECATED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x57)
        timeout: int = field(metadata={'type': uint32})


class MinParentsInCache(ServerMessage):
    """Amount of parents (received through :ref:`PotentialParents`) we should
    keep in cache. Message has not been seen being sent by the server

    :status: DEPRECATED, DEFUNCT
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x58)
        amount: int = field(metadata={'type': uint32})


class DistributedDistributeInterval(ServerMessage):

    """Deprecated distributed network related message

    :status: DEPRECATED, DEFUNCT"""

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x59)
        interval: int = field(metadata={'type': uint32})


class DistributedAliveInterval(ServerMessage):
    """Interval at which a :ref:`DistributedPing` message should be sent to the
    children

    :status: DEPRECATED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x5A)
        interval: int = field(metadata={'type': uint32})


class AddPrivilegedUser(ServerMessage):
    """Usage unknown

    :status: UNKNOWN
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x5B)
        username: str = field(metadata={'type': string})


class CheckPrivileges(ServerMessage):
    """Checks whether the requesting user has privileges, ``time_left`` will
    be ``0`` in case the user has no privileges, time left in seconds otherwise

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x5C)

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x5C)
        time_left: int = field(metadata={'type': uint32})


class ServerSearchRequest(ServerMessage):
    """Search request sent by another user through the server. Upon receiving
    this the message should be passed on to the distributed children

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x5D)
        distributed_code: int = field(metadata={'type': uint8})
        unknown: int = field(metadata={'type': uint32})
        username: str = field(metadata={'type': string})
        ticket: int = field(metadata={'type': uint32})
        query: str = field(metadata={'type': string})


class AcceptChildren(ServerMessage):
    """Tell the server whether or not we are accepting any distributed children,
    the server *should* take this into account when sending
    :ref:`PotentialParents` messages to other peers

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x64)
        accept: bool = field(metadata={'type': boolean})


class PotentialParents(ServerMessage):
    """List of potential parents, used in distributed network

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x66)
        entries: list[PotentialParent] = field(metadata={'type': array, 'subtype': PotentialParent})


class WishlistSearch(ServerMessage):
    """Perform a wishlist search. The interval at which a client should send
    this message is determined by the :ref:`WishlistInterval` message

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x67)
        ticket: int = field(metadata={'type': uint32})
        query: str = field(metadata={'type': string})


class WishlistInterval(ServerMessage):
    """The server lets us know at what interval we should perform wishlist
    searches (:ref:`WishlistSearch`). Sent by the server after logon

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x68)
        interval: int = field(metadata={'type': uint32})


class GetSimilarUsers(ServerMessage):
    """Get a list of similar users

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x6E)

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x6E)
        users: list[SimilarUser] = field(metadata={'type': array, 'subtype': SimilarUser})


class GetItemRecommendations(ServerMessage):
    """Get a list of recommendations based on a single interest

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x6F)
        item: str = field(metadata={'type': string})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x6F)
        item: str = field(metadata={'type': string})
        recommendations: list[Recommendation] = field(metadata={'type': array, 'subtype': Recommendation})


class GetItemSimilarUsers(ServerMessage):
    """Get a list of similar users based on a single interest

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x70)
        item: str = field(metadata={'type': string})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x70)
        item: str = field(metadata={'type': string})
        usernames: list[str] = field(metadata={'type': array, 'subtype': string})


class RoomTickers(ServerMessage):
    """List of chat room tickers (room wall)

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x71)
        room: str = field(metadata={'type': string})
        tickers: list[RoomTicker] = field(metadata={'type': array, 'subtype': RoomTicker})


class RoomTickerAdded(ServerMessage):
    """A ticker has been added to the room (room wall)

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x72)
        room: str = field(metadata={'type': string})
        username: str = field(metadata={'type': string})
        ticker: str = field(metadata={'type': string})


class RoomTickerRemoved(ServerMessage):
    """A ticker has been removed to the room (room wall)

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x73)
        room: str = field(metadata={'type': string})
        username: str = field(metadata={'type': string})


class SetRoomTicker(ServerMessage):
    """Add or update a ticker for a room (room wall)

    .. note::

        An empty ``ticker`` value is not allowed in most clients. However, the server does accept
        it and clears the ticker from the room

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x74)
        room: str = field(metadata={'type': string})
        ticker: str = field(metadata={'type': string})


class AddHatedInterest(ServerMessage):
    """Adds an hated interest. This is used when requesting recommendations
    (eg.: :ref:`GetRecommendations`, ...)

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x75)
        hated_interest: str = field(metadata={'type': string})


class RemoveHatedInterest(ServerMessage):
    """Removes a hated interest previously added with :ref:`AddHatedInterest`
    message

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x76)
        hated_interest: str = field(metadata={'type': string})


class RoomSearch(ServerMessage):
    """Perform a search query on all users in the given room, this can only be
    performed if the room was joined first. The server will send a
    :ref:`FileSearch` to every user in the requested room

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x78)
        room: str = field(metadata={'type': string})
        ticket: int = field(metadata={'type': uint32})
        query: str = field(metadata={'type': string})


class SendUploadSpeed(ServerMessage):
    """Sent to the server right after an upload completed. ``speed`` parameter
    should be in bytes per second. This should not be the global average upload
    speed but rather the upload speed for that particular transfer. After this
    message has been sent the server will recalculate the average speed and
    increase the amount of uploads for your user.

    In exception cases, for example if a transfer was failed midway then
    resumed, only the speed of the resumed part is taken into account. However
    this might be client dependent.

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x79)
        speed: int = field(metadata={'type': uint32})


class GetUserPrivileges(ServerMessage):
    """Retrieve whether a user has privileges

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x7A)
        username: str = field(metadata={'type': string})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x7A)
        username: str = field(metadata={'type': string})
        privileged: bool = field(metadata={'type': boolean})


class GiveUserPrivileges(ServerMessage):
    """Gift a user privileges. This only works if the user sending the message
    has privileges and needs to be less than what the gifting user has left,
    part of its privileges will be taken

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x7B)
        username: str = field(metadata={'type': string})
        days: int = field(metadata={'type': uint32})


class PrivilegesNotification(ServerMessage):

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x7C)
        notification_id: int = field(metadata={'type': uint32})
        username: str = field(metadata={'type': string})


class PrivilegesNotificationAck(ServerMessage):

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x7D)
        notification_id: int = field(metadata={'type': uint32})


class BranchLevel(ServerMessage):
    """Notify the server which branch level we are at in the distributed network

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x7E)
        level: int = field(metadata={'type': uint32})


class BranchRoot(ServerMessage):
    """Notify the server who our branch root user is in the distributed network

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x7F)
        username: str = field(metadata={'type': string})


class ChildDepth(ServerMessage):
    """See :ref:`DistributedChildDepth`

    .. note::

        SoulSeekQt sends the ``depth`` as a ``uint8``

    :status: DEPRECATED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x81)
        depth: int = field(metadata={'type': uint32})


class ResetDistributed(ServerMessage):
    """Server requests to reset our parent and children

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x82)


class PrivateRoomMembers(ServerMessage):
    """List of all members that are part of the private room (excludes owner)

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x85)
        room: str = field(metadata={'type': string})
        usernames: list[str] = field(metadata={'type': array, 'subtype': string})


class PrivateRoomGrantMembership(ServerMessage):
    """Add another user to the private room. Only operators and the owner can
    add members to a private room.

    This message is also received by all other members in the private room

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x86)
        room: str = field(metadata={'type': string})
        username: str = field(metadata={'type': string})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x86)
        room: str = field(metadata={'type': string})
        username: str = field(metadata={'type': string})


class PrivateRoomRevokeMembership(ServerMessage):
    """Remove another user from the private room. Operators can remove regular
    members but not other operators or the owner. The owner can remove anyone
    aside from himself (see :ref:`PrivateRoomDropOwnership`).

    This message is also received by all other members in the private room

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x87)
        room: str = field(metadata={'type': string})
        username: str = field(metadata={'type': string})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x87)
        room: str = field(metadata={'type': string})
        username: str = field(metadata={'type': string})


class PrivateRoomDropMembership(ServerMessage):
    """Drops membership of a private room, this will not do anything for the
    owner of the room. See :ref:`PrivateRoomDropOwnership` for owners

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x88)
        room: str = field(metadata={'type': string})


class PrivateRoomDropOwnership(ServerMessage):
    """Drops ownership of a private room, this disbands the entire room.

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x89)
        room: str = field(metadata={'type': string})


class PrivateRoomMembershipGranted(ServerMessage):
    """Received when the current user has been granted membership to a private
    room

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x8B)
        room: str = field(metadata={'type': string})


class PrivateRoomMembershipRevoked(ServerMessage):
    """Received when the current user had its membership revoked from a private
    room

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x8C)
        room: str = field(metadata={'type': string})


class TogglePrivateRoomInvites(ServerMessage):
    """Enables or disables private room invites (through
    :ref:`PrivateRoomGrantMembership`)

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x8D)
        enable: bool = field(metadata={'type': boolean})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x8D)
        enabled: bool = field(metadata={'type': boolean})


class NewPassword(ServerMessage):
    """Modifies the user's password

    :status: USED"""

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x8E)
        password: str = field(metadata={'type': string})


class PrivateRoomGrantOperator(ServerMessage):
    """Grant operator privileges to a member in a private room. This message
    will also be received by all other members in the room (irrelevant of if
    they are online or not)

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x8F)
        room: str = field(metadata={'type': string})
        username: str = field(metadata={'type': string})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x8F)
        room: str = field(metadata={'type': string})
        username: str = field(metadata={'type': string})


class PrivateRoomRevokeOperator(ServerMessage):
    """Revoke operator privileges from a member in a private room. This message
    will also be received by all other members in the room (irrelevant of if
    they are online or not).

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x90)
        room: str = field(metadata={'type': string})
        username: str = field(metadata={'type': string})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x90)
        room: str = field(metadata={'type': string})
        username: str = field(metadata={'type': string})


class PrivateRoomOperatorGranted(ServerMessage):
    """Received when granted operator privileges in a private room

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x91)
        room: str = field(metadata={'type': string})


class PrivateRoomOperatorRevoked(ServerMessage):
    """Received when operator privileges in a private room were revoked

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x92)
        room: str = field(metadata={'type': string})


class PrivateRoomOperators(ServerMessage):
    """List of operators for a private room

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x94)
        room: str = field(metadata={'type': string})
        usernames: list[str] = field(metadata={'type': array, 'subtype': string})


class PrivateChatMessageUsers(ServerMessage):
    """Send a private message to a list of users. This message will only be
    received by users who have added you using the :ref:`AddUser` message first

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x95)
        usernames: list[str] = field(metadata={'type': array, 'subtype': string})
        message: str = field(metadata={'type': string})


class EnablePublicChat(ServerMessage):
    """Enables public chat, see :ref:`PublicChatMessage`

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x96)


class DisablePublicChat(ServerMessage):
    """Disables public chat, see :ref:`PublicChatMessage`

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x97)


class PublicChatMessage(ServerMessage):
    """When public chat is enabled all messages sent to public rooms will also
    be sent to us using this message. Use :ref:`EnablePublicChat` and
    :ref:`DisablePublicChat` to disable or enable receiving these messages.

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x98)
        room: str = field(metadata={'type': string})
        username: str = field(metadata={'type': string})
        message: str = field(metadata={'type': string})


class GetRelatedSearches(ServerMessage):
    """Usually this is sent by the client right after the :ref:`FileSearch`
    message using the same `query` to retrieve the related searches for that
    query

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x99)
        query: str = field(metadata={'type': string})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x99)
        query: str = field(metadata={'type': string})
        related_searches: list[str] = field(metadata={'type': array, 'subtype': string})


class ExcludedSearchPhrases(ServerMessage):
    """Optionally sent by the server after logging on. Search results containing
    at least one of the phrases (exact match, case insensitive) should be
    filtered out before being sent.

    It is highly recommended to take this filtering into account as not doing so
    could jeopardize the network

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0xA0)
        phrases: list[str] = field(metadata={'type': array, 'subtype': string})


class CannotConnect(ServerMessage):
    """Send to the server if we failed to connect to a peer after a
    :ref:`ConnectToPeer`

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x03E9)
        ticket: int = field(metadata={'type': uint32})
        username: Optional[str] = field(default=None, metadata={'type': string, 'optional': True})

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x03E9)
        ticket: int = field(metadata={'type': uint32})
        username: Optional[str] = field(default=None, metadata={'type': string, 'optional': True})


class CannotCreateRoom(ServerMessage):
    """Sent by the server when attempting to create/join a private room which
    already exists or the user is not part of

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Response(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x03EB)
        room: str = field(metadata={'type': string})


# Peer Initialization messages

class PeerPierceFirewall(PeerInitializationMessage):
    """Sent after connection was successfully established in response to a
    :ref:`ConnectToPeer` message. The ``ticket`` used here should be the ticket
    from that :ref:`ConnectToPeer` message

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint8] = uint8(0x00)
        ticket: int = field(metadata={'type': uint32})


class _PeerInitTicket(uint32):
    # Hacky type here: sometimes people will send a uint64 instead of uint32 as
    # the ticket for a PeerInit message

    @classmethod
    def deserialize(cls, pos: int, data: bytes):
        if len(data[pos:]) == 4:
            return super().deserialize(pos, data)
        else:
            return uint64.deserialize(pos, data)


class PeerInit(PeerInitializationMessage):
    """Sent after direct connection was successfully established (not as a
    response to a :ref:`ConnectToPeer` received from the server)

    The ``ticket`` is usually 0 and was filled in with ``ticket`` value from
    the :ref:`SendConnectTicket` message by older versions of the client

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint8] = uint8(0x01)
        username: str = field(metadata={'type': string})
        typ: str = field(metadata={'type': string})
        ticket: int = field(metadata={'type': _PeerInitTicket})


# Peer messages

class PeerSharesRequest(PeerMessage):
    """Request all shared files/directories from a peer

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x04)
        # Museek docs: PeerSharesReply has an unknown uint32. The assumption is
        # that this is actually a ticket number that was supposed to be passed
        # with this message. The Windows clients appear to accept this ticket
        # number: they will send a reply if a ticket is set but only if this
        # ticket is a uint32, sending the ticket as a uint64 will be rejected.
        # Sending this ticket has no impact on the ticket in PeerSharesReply: it
        # will always be 0
        ticket: Optional[int] = field(default=None, metadata={'type': uint32, 'optional': True})


class PeerSharesReply(PeerMessage):
    """Response to PeerSharesRequest. The response should include empty parent
    directories.

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x05)
        directories: list[DirectoryData] = field(
            metadata={'type': array, 'subtype': DirectoryData}
        )
        # Unknown field that always seems to be 0, possibilities:
        # * This was another list, but it always empty (not tested)
        # * This is a ticket: See explanation of ticket in PeerSharesRequest
        unknown: int = field(default=0, metadata={'type': uint32})
        locked_directories: Optional[list[DirectoryData]] = field(
            default=None,
            metadata={'type': array, 'subtype': DirectoryData, 'optional': True}
        )

        def serialize(self, compress: bool = True) -> bytes:
            return super(type(self), self).serialize(compress)

        @classmethod
        def deserialize(cls, pos: int, message: bytes, decompress: bool = True):
            return super(cls, cls).deserialize(pos, message, decompress)


class PeerSearchReply(PeerMessage):
    """Response to a search request

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x09)
        username: str = field(metadata={'type': string})
        ticket: int = field(metadata={'type': uint32})
        results: list[FileData] = field(metadata={'type': array, 'subtype': FileData})
        has_slots_free: bool = field(metadata={'type': boolean})
        avg_speed: int = field(metadata={'type': uint32})
        # Note: queue_size and unknown. queue_size is described as uint64 in the
        # museek documentation. However I believe that this is actually just a
        # uint32. The other 4 bytes are for an unknown uint32 right before the
        # locked results: the same can be seen in PeerSharesReply
        queue_size: int = field(metadata={'type': uint32})
        unknown: Optional[int] = field(default=0, metadata={'type': uint32})
        locked_results: Optional[list[FileData]] = field(
            default=None,
            metadata={'type': array, 'subtype': FileData, 'optional': True}
        )

        def serialize(self, compress: bool = True) -> bytes:
            return super(type(self), self).serialize(compress)

        @classmethod
        def deserialize(cls, pos: int,  message: bytes, decompress: bool = True):
            return super(cls, cls).deserialize(pos, message, decompress)


class PeerUserInfoRequest(PeerMessage):
    """Request information from the peer

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x0F)


class PeerUserInfoReply(PeerMessage):
    """Response to :ref:`PeerUserInfoRequest`. Possible values for
    ``upload_permissions`` can be found :ref:`here <table-upload-permissions>`

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x10)
        description: str = field(metadata={'type': string})
        has_picture: bool = field(metadata={'type': boolean})
        picture: Optional[bytes] = field(default=None, metadata={'type': bytearr, 'if_true': 'has_picture'})
        upload_slots: int = field(default=0, metadata={'type': uint32})
        queue_size: int = field(default=0, metadata={'type': uint32})
        has_slots_free: bool = field(default=False, metadata={'type': boolean})
        upload_permissions: Optional[int] = field(default=None, metadata={'type': uint32, 'optional': True})


class PeerDirectoryContentsRequest(PeerMessage):
    """Request the contents of a directory

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x24)
        ticket: int = field(metadata={'type': uint32})
        directory: str = field(metadata={'type': string})


class PeerDirectoryContentsReply(PeerMessage):
    """Reply to :ref:`PeerDirectoryContentsRequest`.

    Although the returned directories is an array it will only contain one
    element and will not list files from subdirectories

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x25)
        ticket: int = field(metadata={'type': uint32})
        directory: str = field(metadata={'type': string})
        directories: list[DirectoryData] = field(metadata={'type': array, 'subtype': DirectoryData})

        def serialize(self, compress: bool = True) -> bytes:
            return super(type(self), self).serialize(compress)

        @classmethod
        def deserialize(cls, pos: int, message: bytes, decompress: bool = True):
            return super(cls, cls).deserialize(pos, message, decompress)


class PeerTransferRequest(PeerMessage):
    """``filesize`` can be omitted if the direction==1 however a value of ``0``
    can be used in this case as well

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x28)
        direction: int = field(metadata={'type': uint32})
        ticket: int = field(metadata={'type': uint32})
        filename: str = field(metadata={'type': string})
        filesize: Optional[int] = field(default=None, metadata={'type': uint64, 'optional': True})


class PeerTransferReply(PeerMessage):
    """Response message to :ref:`PeerTransferRequest`

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x29)
        ticket: int = field(metadata={'type': uint32})
        allowed: bool = field(metadata={'type': boolean})
        filesize: Optional[int] = field(default=None, metadata={'type': uint64, 'optional': True, 'if_true': 'allowed'})
        reason: Optional[str] = field(default=None, metadata={'type': string, 'optional': True, 'if_false': 'allowed'})


class PeerTransferQueue(PeerMessage):
    """Request to place the provided transfer of ``filename`` in the queue

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x2B)
        filename: str = field(metadata={'type': string})


class PeerPlaceInQueueReply(PeerMessage):
    """Response to :ref:`PeerPlaceInQueueRequest`

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x2C)
        filename: str = field(metadata={'type': string})
        place: int = field(metadata={'type': uint32})


class PeerUploadFailed(PeerMessage):
    """Sent when uploading failed

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x2E)
        filename: str = field(metadata={'type': string})


class PeerTransferQueueFailed(PeerMessage):
    """Sent when placing the transfer in queue failed

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x32)
        filename: str = field(metadata={'type': string})
        reason: str = field(metadata={'type': string})


class PeerPlaceInQueueRequest(PeerMessage):
    """Request the place of the transfer in the queue.

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x33)
        filename: str = field(metadata={'type': string})


class PeerUploadQueueNotification(PeerMessage):
    """Deprecated message

    :status: DEPRECATED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x34)


# Distributed messages

class DistributedPing(DistributedMessage):
    """Ping request from the parent. Most clients do not send this.

    :status: DEPRECATED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint8] = uint8(0x00)


class DistributedInit(DistributedMessage):
    """Deprecated distributed network related message

    :status: DEPRECATED, DEFUNCT
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint8] = uint8(0x01)
        unknown1: int = field(metadata={'type': uint32})
        unknown2: int = field(metadata={'type': uint32})
        unknown3: str = field(metadata={'type': uint8})
        port: int = field(metadata={'type': uint32})


class DistributedSearchRequest(DistributedMessage):
    """Search request coming from the parent

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint8] = uint8(0x03)
        # Should always be 0x31
        unknown: int = field(metadata={'type': uint32})
        username: str = field(metadata={'type': string})
        ticket: int = field(metadata={'type': uint32})
        query: str = field(metadata={'type': string})


class DistributedBranchLevel(DistributedMessage):
    """Distributed branch level

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint8] = uint8(0x04)
        level: int = field(metadata={'type': uint32})


class DistributedBranchRoot(DistributedMessage):
    """Distributed branch root

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint8] = uint8(0x05)
        username: str = field(metadata={'type': string})


class DistributedChildDepth(DistributedMessage):
    """Used by SoulSeek NS and still passed on by SoulSeekQt, sent to the
    parent upon connecting to that parent (although unclear what happens when
    a peer attaches to another parent while already having children). The parent
    should increase the ``depth`` by 1 until it reaches the branch root, which
    should increase by 1 and send it to the server as a :ref:`ChildDepth`
    message.

    :status: DEPRECATED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint8] = uint8(0x07)
        depth: int = field(metadata={'type': uint32})


class DistributedServerSearchRequest(DistributedMessage):
    """The branch root should just pass the ServerSearchRequest as-is to all its
    children; meaning we will get this message if we are at level 1. If we get
    this message we should translate it to a proper DistributedSearchRequest
    message.

    This message might need to be revisited, as it's only currently used for
    parsing

    :status: USED
    """

    @dataclass(order=True, slots=True)
    class Request(MessageDataclass):
        MESSAGE_ID: ClassVar[uint32] = uint32(0x5D)
        distributed_code: int = field(metadata={'type': uint8})
        unknown: int = field(metadata={'type': uint32})
        username: str = field(metadata={'type': string})
        ticket: int = field(metadata={'type': uint32})
        query: str = field(metadata={'type': string})



from dataclasses import dataclass, field
import enum
import os
from typing import Optional, Union

from ..exceptions import FileNotFoundError
from .utils import normalize_remote_path


class DirectoryShareMode(enum.Enum):
    """Share mode for directories. The mode determines for who the results are
    locked and who can download the file
    """
    EVERYONE = 'everyone'
    FRIENDS = 'friends'
    USERS = 'users'


@dataclass(eq=True, unsafe_hash=True)
class SharedDirectory:
    """Class representing a directory that is explicitly shared by the user.
    This class holds information on how the directory is shared and is a
    container for shared items within this directory.
    """
    directory: str
    absolute_path: str
    alias: str
    share_mode: DirectoryShareMode = field(default=DirectoryShareMode.EVERYONE, compare=False, hash=False)
    users: list[str] = field(default_factory=list, compare=False, hash=False)
    items: set['SharedItem'] = field(default_factory=set, init=False, compare=False, hash=False, repr=False)

    def is_parent_of(self, directory: Union[str, 'SharedDirectory']) -> bool:
        """Returns true if the current directory is any parent of the passed
        shared directory

        :param directory: directory to test, if this directory is of ``str``
            type it should be an absolute path. If it is of
            :class:`.SharedDirectory` type the ``absolute_path`` will be taken
        """
        path = directory if isinstance(directory, str) else directory.absolute_path
        return os.path.commonpath([path, self.absolute_path]) == self.absolute_path

    def is_child_of(self, directory: Union[str, 'SharedDirectory']) -> bool:
        """Returns true if the passed directory is a (sub)child of the current
        directory

        :param directory: directory to test, if this directory is of ``str``
            type it should be an absolute path. If it is of
            :class:`.SharedDirectory` type the ``absolute_path`` will be taken
        """
        path = directory if isinstance(directory, str) else directory.absolute_path
        return os.path.commonpath([self.absolute_path, path]) == path

    def get_remote_path(self) -> str:
        return '@@' + self.alias

    def get_item_by_remote_path(self, remote_path: str) -> 'SharedItem':
        """Returns the :class:`.SharedItem` instance belonging to the passed
        ``remote_path``

        :raise FileNotFoundError: when the item cannot be found in the set of
            `items`
        """
        for item in self.items:
            if item.get_remote_path() == remote_path:
                return item
        else:
            raise FileNotFoundError(
                f"file with remote path {remote_path!r} not found in directory {self!r}")

    def get_items_for_directory(self, directory: 'SharedDirectory') -> set['SharedItem']:
        """Gets items in the current directory that are part of given directory
        """
        return {
            item for item in self.items
            if os.path.commonpath([directory.absolute_path, item.get_absolute_path()]) == directory.absolute_path
        }


@dataclass(eq=True, unsafe_hash=True)
class SharedItem:
    shared_directory: SharedDirectory
    subdir: str
    filename: str
    modified: float
    attributes: Optional[list[tuple[int, int]]] = field(
        default=None,
        init=False,
        compare=False,
        hash=False
    )

    def get_absolute_path(self) -> str:
        """Returns the absolute path of the shared item"""
        return os.path.join(
            self.shared_directory.absolute_path, self.subdir, self.filename)

    def get_remote_path(self) -> str:
        """Returns the full remote path of this file.

        :return: Full remote file path: Example: "@@abcde\\Music\\song.mp3"
        """
        return normalize_remote_path(
            '@@' + os.path.join(self.shared_directory.alias, self.subdir, self.filename))

    def get_remote_directory_path(self) -> str:
        """Returns the remote directory path this file resides in

        :return: Remote directory path. Example: "@@abcde\\Music"
        """
        return normalize_remote_path(
            '@@' + os.path.join(self.shared_directory.alias, self.subdir))

    def get_remote_directory_path_parts(self) -> tuple[str, ...]:
        """Returns the remote directory path split into parts"""
        return tuple(self.get_remote_directory_path().split('\\'))

    def get_query_path(self) -> str:
        """Returns the query-able part of the shared item

        :return: Queryable path of the item. Example: "Music\\Metal\\Song.mp3"
        """
        return normalize_remote_path(os.path.join(self.subdir, self.filename))

    def __getstate__(self):
        fields = self.__dict__.copy()
        fields['shared_directory'] = None
        return fields



from __future__ import annotations
import aiofiles
from aiofiles import os as asyncos
import asyncio
from async_timeout import timeout as atimeout
from enum import auto, Flag
import logging
from operator import itemgetter
import os
import time
from typing import Optional, TYPE_CHECKING

from ..base_manager import BaseManager
from .cache import TransferNullCache, TransferCache
from ..constants import (
    MAX_TRANSFER_MGMT_INTERVAL,
    MIN_TRANSFER_MGMT_INTERVAL,
    TRANSFER_REPLY_TIMEOUT,
)
from ..exceptions import (
    AioSlskException,
    ConnectionReadError,
    ConnectionWriteError,
    FileNotFoundError,
    FileNotSharedError,
    InvalidStateTransition,
    PeerConnectionError,
    RequestPlaceFailedError,
    TransferNotFoundError,
)
from ..network.connection import (
    CloseReason,
    PeerConnection,
    PeerConnectionState,
    PeerConnectionType,
)
from ..events import (
    build_message_map,
    on_message,
    BlockListChangedEvent,
    EventBus,
    FriendListChangedEvent,
    MessageReceivedEvent,
    PeerInitializedEvent,
    SessionInitializedEvent,
    ScanCompleteEvent,
    SharedDirectoryChangeEvent,
    TransferAddedEvent,
    TransferProgressEvent,
    TransferRemovedEvent,
)
from ..protocol.primitives import uint32, uint64
from ..protocol.messages import (
    AddUser,
    GetUserStatus,
    PeerPlaceInQueueReply,
    PeerPlaceInQueueRequest,
    PeerTransferQueue,
    PeerTransferQueueFailed,
    PeerTransferReply,
    PeerTransferRequest,
    PeerUploadFailed,
    SendUploadSpeed,
)
from .model import AbortReason, FailReason, Transfer, TransferDirection
from .state import TransferState
from ..settings import Settings
from ..shares.manager import SharesManager
from ..tasks import BackgroundTask
from ..user.manager import UserManager
from ..user.model import BlockingFlag, UserStatus, TrackingFlag
from ..utils import task_counter, ticket_generator

if TYPE_CHECKING:
    from ..network.network import Network


logger = logging.getLogger(__name__)


class _RequestFlag(Flag):
    SHARES_CHANGE = auto()
    TRANSFER_CHANGE = auto()


class TransferManager(BaseManager):
    """Class responsible for transfer related functionality. This class stores
    transfer objects and handles related events.
    """

    def __init__(
            self, settings: Settings, event_bus: EventBus,
            user_manager: UserManager, shares_manager: SharesManager,
            network: Network, cache: Optional[TransferCache] = None):

        self.cache: TransferCache = cache if cache else TransferNullCache()
        self._settings: Settings = settings
        self._event_bus: EventBus = event_bus
        self._user_manager: UserManager = user_manager
        self._shares_manager: SharesManager = shares_manager
        self._network: Network = network

        self._ticket_generator = ticket_generator()

        self._transfers: list[Transfer] = []
        self._file_connection_futures: dict[int, asyncio.Future] = {}
        self._progress_reporting_task: BackgroundTask = BackgroundTask(
            interval=self._settings.transfers.report_interval,
            task_coro=self._progress_reporting_job,
            name='transfer-progress-task'
        )

        self._management_queue: asyncio.Queue = asyncio.Queue(maxsize=1)
        self._management_task: BackgroundTask = BackgroundTask(
            interval=MIN_TRANSFER_MGMT_INTERVAL,
            task_coro=self._management_job,
            name='transfer-management-task'
        )
        self._management_lock: asyncio.Lock = asyncio.Lock()
        self._management_flags: _RequestFlag = _RequestFlag(0)

        self._MESSAGE_MAP = build_message_map(self)

        self.register_listeners()

    @property
    def transfers(self):
        return self._transfers

    def register_listeners(self):
        self._event_bus.register(
            MessageReceivedEvent, self._on_message_received)
        self._event_bus.register(
            PeerInitializedEvent, self._on_peer_initialized)
        self._event_bus.register(
            SessionInitializedEvent, self._on_session_initialized)

        self._event_bus.register(
            BlockListChangedEvent, self._request_shares_cycle)
        self._event_bus.register(
            FriendListChangedEvent, self._request_shares_cycle)
        self._event_bus.register(
            SharedDirectoryChangeEvent, self._request_shares_cycle)
        self._event_bus.register(
            ScanCompleteEvent, self._request_shares_cycle)

    async def read_cache(self):
        """Reads the transfers from the caches and corrects the state of those
        transfers
        """
        transfers: list[Transfer] = self.cache.read()
        for transfer in transfers:
            # Analyze the current state of the stored transfers and set them to
            # the correct state
            transfer.remotely_queued = False
            if transfer.state.VALUE == TransferState.INITIALIZING:
                await transfer.state.queue()

            elif transfer.is_transferring():
                if transfer.is_transfered():
                    state = TransferState.COMPLETE
                else:
                    state = TransferState.INCOMPLETE
                transfer.state = TransferState.init_from_state(state, transfer)
                transfer.reset_time_vars()

            await self.add(transfer)

    def write_cache(self):
        """Write all currently stored transfers to the cache"""
        self.cache.write(self._transfers)

    async def load_data(self):
        await self.read_cache()

    async def store_data(self):
        self.write_cache()

    async def start(self):
        self._management_task.start()
        await self.start_progress_reporting_task()

    async def stop(self) -> list[asyncio.Task]:
        """Cancel all current transfer tasks

        :return: a list of tasks that have been cancelled so that they can be
            awaited
        """
        cancelled_tasks = []
        for transfer in self.transfers:
            cancelled_tasks.extend(transfer.cancel_tasks())

        if task := self._progress_reporting_task.cancel():
            cancelled_tasks.append(task)

        if task := self._management_task.cancel():
            cancelled_tasks.append(task)

        self._management_queue = asyncio.Queue()

        return cancelled_tasks

    async def start_progress_reporting_task(self):
        """Start the transfer progress reporting background task which will emit
        :class:`aioslsk.events.TransferProgressEvent` events at an interval
        defined in the settings. This method is called automatically when
        starting the client
        """
        self._progress_reporting_task.start()

    def stop_progress_reporting_task(self):
        """Start the transfer progress reporting task"""
        self._progress_reporting_task.cancel()

    async def _progress_reporting_job(self):
        updates = []
        for transfer in self._transfers:
            previous = transfer.progress_snapshot
            current = transfer.take_progress_snapshot()
            if previous != current:
                updates.append((transfer, previous, current))

        if updates:
            await self._event_bus.emit(TransferProgressEvent(updates))

    async def download(self, username: str, filename: str, paused: bool = False) -> Transfer:
        """Requests to start a downloading the file from the given user

        :param user: User from which to download the file
        :param filename: Name of the file to download. This should be the full
            path to the file as returned in the search results
        :param paused: Adds the download in the paused state
        :return: a :class:`.Transfer` object from which the status of the
            transfer can be requested. If the transfer already exists in the
            client then this transfer will be returned
        """
        transfer = await self.add(
            Transfer(
                username,
                filename,
                TransferDirection.DOWNLOAD
            )
        )
        if paused:
            await transfer.state.pause()
        else:
            await transfer.state.queue()

        return transfer

    async def abort(self, transfer: Transfer):
        """Aborts the given transfer. This will cancel all pending transfers
        and remove the file (in case of download)

        :param transfer: :class:`.Transfer` object to abort
        :raise TransferNotFoundError: if the transfer has not been added to the
            manager first
        :raise InvalidStateTransition: When the transfer could not be
            transitioned to aborted
        """
        if transfer not in self.transfers:
            raise TransferNotFoundError(
                "cannot queue transfer: transfer was not added to the manager")

        if not await transfer.state.abort(reason=AbortReason.REQUESTED):
            raise InvalidStateTransition(
                transfer,
                transfer.state.VALUE,
                TransferState.State.ABORTED,
                "Could not make the desired state transition"
            )

    async def queue(self, transfer: Transfer):
        """Places given transfer (back) in the queue

        This method can be called on downloads that are in the following states:

        * ``ABORTED``: Re-starts the aborted download
        * ``PAUSED``: Continues the paused download
        * ``COMPLETE``: Re-downloads the file. By default this will download
          the file to a new location unless the initial file was removed
        * ``INCOMPLETE``: The library should automatically attempt to retry
          a download in this state
        * ``FAILED``: Re-attempt to download the file

        :param transfer: :class:`.Transfer` object to queue
        :raise TransferNotFoundError: if the transfer has not been added to the
            manager first
        :raise InvalidStateTransition: When the transfer could not be
            transitioned to queued
        """
        if transfer not in self.transfers:
            raise TransferNotFoundError(
                "cannot queue transfer: transfer was not added to the manager")

        has_transitioned = await transfer.state.queue()
        if not has_transitioned:
            raise InvalidStateTransition(
                transfer,
                transfer.state.VALUE,
                TransferState.State.QUEUED,
                "Could not make the desired state transition"
            )

    async def pause(self, transfer: Transfer):
        if transfer not in self.transfers:
            raise TransferNotFoundError(
                "cannot pause transfer: transfer was not added to the manager")

        has_transitioned = await transfer.state.pause()
        if not has_transitioned:
            raise InvalidStateTransition(
                transfer,
                transfer.state.VALUE,
                TransferState.State.PAUSED,
                "Could not make the desired state transition"
            )

    async def add(self, transfer: Transfer) -> Transfer:
        """Adds a transfer if it does not already exist, otherwise it returns
        the already existing transfer. This method will emit a
        :class:`~aioslsk.events.TransferAddedEvent` only if the transfer did not
        exist.

        This method only adds a transfer and does not automatically start it. To
        do so call :func:`queue` on the manager

        :param transfer: Transfer to be added
        :return: either the transfer we have passed or the already existing
            transfer
        """
        for queued_transfer in self._transfers:
            if queued_transfer == transfer:
                logger.info("skip adding transfer, already exists : %s", queued_transfer)
                return queued_transfer

        logger.info("adding transfer : %s", transfer)
        transfer.state_listeners.append(self)
        self._transfers.append(transfer)

        self.request_management_cycle(_RequestFlag.TRANSFER_CHANGE)
        await self._event_bus.emit(TransferAddedEvent(transfer))

        return transfer

    async def remove(self, transfer: Transfer):
        """Remove a transfer from the list of transfers. This will attempt to
        abort the transfer before removing it. Emits a
        :class:`.TransferRemovedEvent` after removal

        :param transfer: Transfer object to remove
        :raise TransferNotFoundError: Raised when the transfer was not added to
            the manager
        """
        if transfer not in self.transfers:
            raise TransferNotFoundError(
                "cannot remove transfer: transfer was not added to the manager")

        try:
            await self.abort(transfer)
        except InvalidStateTransition:
            pass
        except Exception:
            logger.exception("error aborting transfer before removal : %s", transfer)
        finally:
            self._transfers.remove(transfer)
            await self._event_bus.emit(TransferRemovedEvent(transfer))

        self.request_management_cycle(_RequestFlag.TRANSFER_CHANGE)

    def get_uploads(self) -> list[Transfer]:
        return [transfer for transfer in self._transfers if transfer.is_upload()]

    def get_downloads(self) -> list[Transfer]:
        return [transfer for transfer in self._transfers if transfer.is_download()]

    def get_upload_slots(self) -> int:
        """Returns the total amount of upload slots"""
        return self._settings.transfers.limits.upload_slots

    def has_slots_free(self) -> bool:
        return self.get_free_upload_slots() > 0

    def get_free_upload_slots(self) -> int:
        uploading_transfers = self.get_uploading()
        available_slots = self.get_upload_slots() - len(uploading_transfers)
        return max(0, available_slots)

    def get_queue_size(self) -> int:
        """Returns the amount of queued uploads"""
        return len([
            transfer for transfer in self._transfers
            if transfer.is_upload() and transfer.state == TransferState.QUEUED
        ])

    def get_downloading(self) -> list[Transfer]:
        """Returns all transfers that are currently downloading or an attempt is
        is made to start downloading
        """
        return [
            transfer for transfer in self._transfers
            if transfer.is_download() and transfer.is_processing()
        ]

    def get_uploading(self) -> list[Transfer]:
        """Returns all transfers that are currently uploading or an attempt is
        is made to start uploading
        """
        return [
            transfer for transfer in self._transfers
            if transfer.is_upload() and transfer.is_processing()
        ]

    def get_finished_transfers(self) -> list[Transfer]:
        """Returns a complete list of transfers that are in a finalized state
        (COMPLETE, ABORTED, FAILED)
        """
        return [
            transfer for transfer in self._transfers
            if transfer.is_finalized()
        ]

    def get_unfinished_transfers(self) -> list[Transfer]:
        """Returns a complete list of transfers that are not in a finalized
        state (COMPLETE, ABORTED, FAILED)
        """
        return [
            transfer for transfer in self._transfers
            if not transfer.is_finalized()
        ]

    def get_download_speed(self) -> float:
        """Return current download speed (in bytes/second)"""
        return sum(transfer.get_speed() for transfer in self.get_downloading())

    def get_upload_speed(self) -> float:
        """Return current upload speed (in bytes/second)"""
        return sum(transfer.get_speed() for transfer in self.get_uploading())

    def get_average_upload_speed(self) -> float:
        """Returns average upload speed (in bytes/second)"""
        upload_speeds = [
            transfer.get_speed() for transfer in self._transfers
            if transfer.state.VALUE == TransferState.COMPLETE and transfer.is_upload()
        ]
        if len(upload_speeds) == 0:
            return 0.0
        return sum(upload_speeds) / len(upload_speeds)

    def get_place_in_queue(self, transfer: Transfer) -> int:
        """Gets the place of the given upload in the transfer queue

        :return: The place in the queue, 0 if not in the queue a value equal or
            greater than 1 indicating the position otherwise
        """
        _, uploads = self._get_queued_transfers()
        try:
            return uploads.index(transfer)
        except ValueError:
            return 0

    def get_transfer(self, username: str, remote_path: str, direction: TransferDirection) -> Transfer:
        """Lookup transfer by ``username``, ``remote_path`` and ``transfer``
        direction. This method will raise an exception in case the transfer is
        not found

        :param username: Username of the transfer
        :param remote_path: Full remote path of the transfer
        :param direction: Direction of the transfer (upload / download)
        :return: The matching transfer object
        :raise ValueError: If the transfer is not found
        """
        req_transfer = Transfer(username, remote_path, direction)
        for transfer in self._transfers:
            if transfer == req_transfer:
                return transfer

        raise ValueError(
            f"transfer for user {username} and remote_path {remote_path} (direction={direction}) not found")

    def find_transfer(self, username: str, remote_path: str, direction: TransferDirection) -> Optional[Transfer]:
        """Lookup transfer by ``username``, ``remote_path`` and ``transfer``
        direction. This method will return ``None`` if the transfer is not found

        :param username: Username of the transfer
        :param remote_path: Full remote path of the transfer
        :param direction: Direction of the transfer (upload / download)
        :return: The matching transfer object or ``None``
        """
        req_transfer = Transfer(username, remote_path, direction)
        for transfer in self._transfers:
            if transfer == req_transfer:
                return transfer

        return None

    async def manage_user_tracking(self):
        """Remove or add user tracking based on the list of transfers. This
        method will untrack users for which there are no more unfinished
        transfers and start/keep tracking users for which there are unfinished
        transfers
        """
        unfinished_users = set(
            transfer.username
            for transfer in self.get_unfinished_transfers()
        )
        finished_users = set(
            transfer.username
            for transfer in self.get_finished_transfers()
        )

        for username in unfinished_users:
            await self._user_manager.track_user(username, TrackingFlag.TRANSFER)
        for username in finished_users - unfinished_users:
            await self._user_manager.untrack_user(username, TrackingFlag.TRANSFER)

    async def _management_job(self) -> float:
        await self._management_queue.get()

        start = time.monotonic()

        flags = self._management_flags
        self._management_flags = _RequestFlag(0)

        if flags & _RequestFlag.SHARES_CHANGE:
            await self.manage_shares_changed()

        await self.manage_user_tracking()
        self.manage_transfers()

        duration = time.monotonic() - start

        return min(MIN_TRANSFER_MGMT_INTERVAL + duration, MAX_TRANSFER_MGMT_INTERVAL)

    def request_management_cycle(self, flag: _RequestFlag):
        self._management_flags |= flag
        try:
            self._management_queue.put_nowait(None)
        except asyncio.QueueFull:
            pass

    def manage_transfers(self):
        """This method analyzes the state of the current downloads/uploads and
        starts them up in case there are free slots available
        """
        downloads, uploads = self._get_queued_transfers()
        free_upload_slots = self.get_free_upload_slots()

        # Downloads will just get remotely queued
        for download in downloads:
            download._remotely_queue_task = asyncio.create_task(
                self._queue_remotely(download),
                name=f'queue-remotely-{task_counter()}'
            )
            download._remotely_queue_task.add_done_callback(
                download._remotely_queue_task_complete
            )

        # Uploads should be initialized and uploaded if possible
        for upload in uploads[:free_upload_slots]:
            upload._transfer_task = asyncio.create_task(
                self._initialize_upload(upload),
                name=f'initialize-upload-{task_counter()}'
            )
            upload._transfer_task.add_done_callback(
                upload._transfer_task_complete
            )

    async def manage_shares_changed(self):
        logger.debug("processing shares or block list changes")

        uploads = filter(
            lambda transfer: (
                transfer.is_upload()
                and transfer.state.VALUE not in (TransferState.COMPLETE, TransferState.FAILED)
            ),
            self.transfers
        )

        tasks = []
        for upload in uploads:

            should_change, abort_reason = self._evaluate_aborted_state(upload)

            if should_change:
                if upload.state.VALUE == TransferState.ABORTED:
                    tasks.append(upload.state.queue())
                else:
                    tasks.append(upload.state.abort(reason=abort_reason))

            elif abort_reason:
                upload.abort_reason = abort_reason

        await asyncio.gather(*tasks, return_exceptions=True)

    def _get_queued_transfers(self) -> tuple[list[Transfer], list[Transfer]]:
        """Returns all transfers eligable for being initialized

        :return: a tuple containing 2 lists: the eligable downloads and eligable
            uploads
        """
        uploading_users: set[str] = {
            transfer.username for transfer in self._transfers
            if transfer.is_upload() and transfer.is_processing()
        }
        users_with_queued_upload: set[str] = set()

        queued_downloads: list[Transfer] = []
        queued_uploads: list[Transfer] = []
        for transfer in self._transfers:
            # Get the user object from the user manager, if the user is tracked
            # this user object will be returned. Otherwise a new user object is
            # created, but not assigned to the user manager, whose status is
            # UNKNOWN, giving it the benefit of the doubt and scheduling the
            # transfer
            user = self._user_manager.get_user_object(transfer.username)
            if user.status == UserStatus.OFFLINE:
                continue

            if transfer.direction == TransferDirection.UPLOAD:
                # Do not add the user if the user is already uploading or a
                # transfer was already added to the queud upload list (only
                # allow 1 transfer per user)
                if transfer.username in uploading_users:
                    continue
                if transfer.username in users_with_queued_upload:
                    continue

                if transfer.state.VALUE == TransferState.QUEUED:
                    users_with_queued_upload.add(transfer.username)
                    queued_uploads.append(transfer)

            else:
                if transfer.remotely_queued:
                    continue

                # For downloads we try to continue with incomplete downloads,
                # for uploads it's up to the other user
                # Failed downloads without a reason are retried
                if transfer.state.VALUE in (TransferState.QUEUED, TransferState.INCOMPLETE):
                    queued_downloads.append(transfer)
                elif transfer.state.VALUE == TransferState.FAILED and transfer.fail_reason is None:
                    queued_downloads.append(transfer)

        queued_uploads = self._prioritize_uploads(queued_uploads)

        return queued_downloads, queued_uploads

    def _prioritize_uploads(self, uploads: list[Transfer]) -> list[Transfer]:
        """Ranks the queued uploads by priority based on certain parameters

        :return: sorted list of provided by uploads by priority
        """
        friends = self._settings.users.friends

        ranking = []
        for upload in uploads:
            user = self._user_manager.get_user_object(upload.username)

            rank = 0
            # Rank UNKNOWN status lower, OFFLINE should be blocked
            if user.status in (UserStatus.ONLINE, UserStatus.AWAY):
                rank += 1

            if upload.username in friends:
                rank += 5

            # Privileged users should have absolute priority
            if user.privileged:
                rank += 100

            ranking.append((rank, upload))

        ranking.sort(key=itemgetter(0))
        return list(reversed([upload for _, upload in ranking]))

    async def _prepare_download_path(self, transfer: Transfer):
        if transfer.local_path is None:
            download_path, file_path = self._shares_manager.calculate_download_path(transfer.remote_path)
            transfer.local_path = os.path.join(download_path, file_path)

        path, _ = os.path.split(transfer.local_path)
        await self._shares_manager.create_directory(path)

    async def _calculate_offset(self, transfer: Transfer) -> int:
        """Calculates the offset when downloading a file by inspecting the file
        at the local path.

        :return: the calculated offset (in bytes)
        """
        # Shouldn't occur but this is to keep the typing happy
        if not transfer.local_path:
            return 0

        try:
            return await asyncos.path.getsize(transfer.local_path)
        except (OSError, TypeError):
            return 0

    def _reset_remotely_queued_flags(self, username: str):
        """Removes the ``remotely_queued`` flag from all downloads for the given
        given user
        """
        for transfer in self._transfers:
            if transfer.direction == TransferDirection.DOWNLOAD:
                if transfer.username == username:
                    transfer.remotely_queued = False

    async def _queue_remotely(self, transfer: Transfer):
        """Remotely queue the given transfer. If the message was successfully
        delivered the ``remotely_queued`` flag will be set for the transfer
        """
        logger.debug("attempting to queue transfer remotely : %s", transfer)
        try:
            await self._network.send_peer_messages(
                transfer.username,
                PeerTransferQueue.Request(transfer.remote_path)
            )

        except (ConnectionWriteError, PeerConnectionError) as exc:
            logger.debug("failed to queue transfer remotely : %s : %r", transfer, exc)
            transfer.increase_queue_attempts()
            await transfer.state.queue()

        else:
            transfer.remotely_queued = True
            transfer.reset_queue_attempts()
            self.request_management_cycle(_RequestFlag.TRANSFER_CHANGE)

    async def request_place_in_queue(self, transfer: Transfer) -> Optional[int]:
        """Requests the place in queue for the given transfer. The method will
        return the value in case of success and return the value

        :return: place in queue or ``None`` in case of error
        :raise RequestPlaceFailedError: when the request failed to send to the
            peer or waiting for a response timed out
        """
        try:
            await self._network.send_peer_messages(
                transfer.username,
                PeerPlaceInQueueRequest.Request(transfer.remote_path)
            )
        except ConnectionWriteError:
            logger.info("failed to request place in queue of transfer : %s", transfer)
            raise RequestPlaceFailedError("failed to request place in queue")

        try:
            async with atimeout(15):
                _, response = await self._network.create_peer_response_future(
                    peer=transfer.username,
                    message_class=PeerPlaceInQueueReply.Request,
                    fields={
                        'filename': transfer.remote_path
                    }
                )
        except asyncio.TimeoutError:
            logger.warning(
                "timeout receiving response to place in queue for transfer : %s", transfer)
            raise RequestPlaceFailedError("failed to request place in queue")

        else:
            transfer.place_in_queue = response.place
            return response.place

    async def _initialize_download(
            self, transfer: Transfer, peer_connection: PeerConnection,
            request: PeerTransferRequest.Request):
        """Initializes a download and starts downloading. This method should be
        called after a :class:`.PeerTransferRequest` has been received

        Following steps are taken in this method:

        1. Set the state of the transfer to INITIALIZING
        2. Send a :class:`PeerTransferReply` with allowed set to true

           * In case the message cannot be delivered put the transfer back to
             QUEUED

        3. Create a future for an incoming file connection with the ticket from
           the :class:`PeerTransferRequest` message, await the future

           * In case a timeout occurs waiting for the file connection: put the
             transfer to QUEUED

        4. Calculate the transfer offset and send it

           * In case this fails: put transfer to INCOMPLETE

        5. Start downloading, see :meth:`_download_file` : exception cases are
           handled internally by this method

        :param transfer: transfer object to initialize
        :param peer_connection: peer connection on which the transfer request
            was received
        :param request: transfer request object for the given transfer
        """
        await transfer.state.initialize()

        transfer.filesize = request.filesize

        try:
            await peer_connection.send_message(
                PeerTransferReply.Request(
                    ticket=request.ticket,
                    allowed=True
                )
            )
        except ConnectionWriteError:
            logger.warning(
                "failed to send transfer reply for ticket %d and transfer : %s", request.ticket, transfer)
            await transfer.state.queue()
            return

        # Already create a future for the incoming connection
        file_connection_future: asyncio.Future = asyncio.Future()
        self._file_connection_futures[request.ticket] = file_connection_future

        try:
            async with atimeout(60):
                file_connection: PeerConnection = await file_connection_future

        except asyncio.TimeoutError:
            await transfer.state.queue()
            return

        # The transfer ticket should already have been received (otherwise the
        # future could not have completed)
        # Calculate and send the file offset
        offset = await self._calculate_offset(transfer)
        transfer.bytes_transfered = offset
        try:
            await file_connection.send_message(uint64(offset).serialize())

        except ConnectionWriteError:
            logger.warning("failed to send offset : %d : %s", offset, transfer)
            if transfer.is_upload():
                await transfer.state.fail()
            else:
                await transfer.state.incomplete()
            return

        except asyncio.CancelledError:
            # Aborted or program shut down
            logger.debug("requested to cancel transfer: %s", transfer)
            await file_connection.disconnect(CloseReason.REQUESTED)
            raise

        if transfer.direction == TransferDirection.DOWNLOAD:
            await self._download_file(transfer, file_connection)
        else:
            await self._upload_file(transfer, file_connection)

    async def _initialize_upload(self, transfer: Transfer):
        """Attempts to initialize and upload to another peer. Following steps
        are taken in this method:

        1. Set the state of the transfer to INITIALIZING
        2. Notify the downloader we are ready to upload a file by sending a
           :class:`.PeerTransferRequest` message, this includes a ticket

           * In case the message cannot be delivered put the transfer back to
             QUEUED

        3. Wait for a response in the form of a :class:`PeerTransferReply`
           message

           * In case no response in received on time: put transfer to QUEUED
           * In case the response message disallows upload: put the transfer to
             FAILED

        4. Open a file connection

           * In case this fails: put transfer to QUEUED

        5. Send the transfer ticket

           * In case this fails: put transfer to QUEUED

        6. Wait for transfer offset

           * In case this fails: put transfer to QUEUED

        7. Start uploading, see :meth:`_upload_file` : exception cases are
           handled internally by this method
        8. Report upload speed to the server : :class:`SendUploadSpeed`
        """
        await transfer.state.initialize()

        ticket = next(self._ticket_generator)

        try:
            await self._network.send_peer_messages(
                transfer.username,
                PeerTransferRequest.Request(
                    TransferDirection.DOWNLOAD.value,
                    ticket,
                    transfer.remote_path,
                    filesize=transfer.filesize
                )
            )

        except (ConnectionWriteError, PeerConnectionError) as exc:
            logger.debug("failed to send request to upload : %s : %r", transfer, exc)
            await transfer.state.queue()
            return

        try:
            async with atimeout(TRANSFER_REPLY_TIMEOUT):
                connection, response = await self._network.create_peer_response_future(
                    transfer.username,
                    PeerTransferReply.Request,
                    fields={
                        'ticket': ticket
                    }
                )

        except asyncio.TimeoutError:
            logger.debug("timeout waiting for transfer reply : %s", transfer)
            await transfer.state.queue()
            return

        if not response.allowed:
            await transfer.state.fail(reason=response.reason)
            return

        # Create a file connection
        try:
            connection = await self._network.create_peer_connection(
                transfer.username,
                PeerConnectionType.FILE
            )

        except PeerConnectionError:
            logger.info("failed to create peer connection for transfer : %s", transfer)
            await transfer.state.queue()
            return

        # Send transfer ticket
        try:
            await connection.send_message(uint32(ticket).serialize())
        except ConnectionWriteError:
            logger.info("failed to send transfer ticket : %d : %s", ticket, transfer)
            await transfer.state.queue()
            return

        # Receive transfer offset
        try:
            transfer.bytes_transfered = await connection.receive_transfer_offset()
        except ConnectionReadError:
            logger.info("failed to receive transfer offset : %s", transfer)
            await transfer.state.queue()
            return

        else:
            logger.debug("received offset for transfer : %d : %s", transfer.bytes_transfered, transfer)

        await self._upload_file(transfer, connection)

        # Send transfer speed
        if transfer.state.VALUE == TransferState.COMPLETE:
            self._network.queue_server_messages(
                SendUploadSpeed.Request(int(transfer.get_speed()))
            )

    async def _upload_file(self, transfer: Transfer, connection: PeerConnection):
        """Uploads the transfer over the connection. This method will set the
        appropriate states on the passed ``transfer`` and ``connection`` objects

        This method performs the following actions:

        1. Set the state of the transfer to TRANSFERRING
        2. Opens the upload file and set the file pointer to transfer offset

           * In case the file cannot be opened the transfer goes to FAILED

        3. Starts transfering the file

           * In case the local path is not set on the file :

             * Put the transfer to FAILED
             * Disconnect the socket

           * In case an error occurs while reading the file :

             * Put the transfer to FAILED
             * Disconnect the socket

           * In case a network error occurs :

             * Put the transfer to FAILED
             * Send a :class:`PeerUploadFailed` message to the peer

           * In case this task is cancelled :

             * Disconnect the socket

        4. If the file is fully transferred, wait for the other peer to close
           the socket
        5. Verify the that file has been completely transfered

           * In case it is : COMPLETE
           * In case it is not : FAILED

        :param transfer: :class:`.Transfer` object
        :param connection: connection on which file should be sent
        """
        connection.set_connection_state(PeerConnectionState.TRANSFERRING)
        await transfer.state.start_transferring()

        try:
            if not transfer.local_path:
                raise AioSlskException(
                    f"attempted to upload a transfer that doesn't have a local path set : {transfer}")

            async with aiofiles.open(transfer.local_path, mode='rb') as handle:
                await handle.seek(transfer.bytes_transfered)
                await connection.send_file(
                    handle,
                    transfer._transfer_progress_callback
                )

        except OSError:
            logger.exception("error opening local file : %s", transfer.local_path)
            await transfer.state.fail(reason=FailReason.FILE_READ_ERROR)
            await connection.disconnect(CloseReason.REQUESTED)

        except ConnectionWriteError:
            logger.warning("error writing to socket for transfer : %s", transfer)
            await transfer.state.fail()

            # Possible this needs to be put in a task or just queued, if the
            # peer went offline it's possible we are hogging the _current_task
            # for nothing (sending the peer message could time out)
            try:
                await self._network.send_peer_messages(
                    transfer.username,
                    PeerUploadFailed.Request(transfer.remote_path)
                )

            except PeerConnectionError:
                logger.info("failed to send PeerUploadFailed message (possibly peer went offline)")

        except asyncio.CancelledError:
            # Aborted or program shut down
            await connection.disconnect(CloseReason.REQUESTED)
            raise

        except AioSlskException:
            logger.exception("failed to upload transfer : %s", transfer)
            await transfer.state.fail(FailReason.FILE_NOT_SHARED)

            await connection.disconnect(CloseReason.REQUESTED)

            try:
                await self._network.send_peer_messages(
                    transfer.username,
                    PeerUploadFailed.Request(transfer.remote_path)
                )

            except PeerConnectionError:
                logger.debug("failed to send PeerUploadFailed message (possibly peer went offline)")

        else:
            await connection.receive_until_eof(raise_exception=False)
            if transfer.is_transfered():
                await transfer.state.complete()
            else:
                await transfer.state.fail()

    async def _download_file(self, transfer: Transfer, connection: PeerConnection):
        """Downloads the transfer over the connection. This method will set the
        appropriate states on the passed ``transfer`` and ``connection`` objects.

        This method will:

        1. Calculate and create the download path

           * In case of failure fail the download and disconnect

        2. Set the state of the transfer to TRANSFERRING
        3. Open and start receiving the file, the amount of bytes that are
           expected to be received will be calculated from the offset

           * In case of failure to write to the file: put the transfer to FAILED
           * In case a read error occurred on the socket: put the transfer to INCOMPLETE

        4. Verify the transfer is complete

           * If all bytes are transfered: put the transfer to COMPLETE
           * If not all bytes are transfered: put the transfer to FAIL

        :param transfer: :class:`.Transfer` object of the download
        :param connection: connection on which file should be received
        """
        try:
            await self._prepare_download_path(transfer)
        except OSError:
            logger.exception("failed to create path : %s", transfer.local_path)
            await connection.disconnect(CloseReason.REQUESTED)
            await transfer.state.fail(reason=FailReason.FILE_READ_ERROR)
            return

        connection.set_connection_state(PeerConnectionState.TRANSFERRING)
        await transfer.state.start_transferring()

        if transfer.local_path is None:  # pragma: no cover
            raise AioSlskException(
                f"attempted to start download for which local_path was not set : {transfer}")

        if transfer.filesize is None:  # pragma: no cover
            raise AioSlskException(
                f"attempted to start download for which filesize was not set : {transfer}")

        try:
            async with aiofiles.open(transfer.local_path, mode='ab') as handle:
                await connection.receive_file(
                    handle,
                    transfer.filesize - transfer.bytes_transfered,
                    transfer._transfer_progress_callback
                )

        except OSError:
            logger.exception("error opening local file : %s", transfer.local_path)
            await transfer.state.fail(reason=FailReason.FILE_READ_ERROR)

            await connection.disconnect(CloseReason.REQUESTED)

        except ConnectionReadError as exc:
            logger.warning("error reading from socket : %s", transfer, exc_info=exc)
            await transfer.state.incomplete()

        except asyncio.CancelledError:
            # Aborted or program shut down. The state does not need to be set
            # here as this usually occurs during a state transition (aborting)
            logger.debug("requested to cancel transfer: %s", transfer)
            await connection.disconnect(CloseReason.REQUESTED)
            raise

        else:
            await connection.disconnect(CloseReason.REQUESTED)
            if transfer.is_transfered():
                await transfer.state.complete()
            else:
                await transfer.state.fail(reason=FailReason.CANCELLED)

    async def on_transfer_state_changed(
            self, transfer: Transfer, old: TransferState.State, new: TransferState.State):

        self.request_management_cycle(_RequestFlag.TRANSFER_CHANGE)

    async def _on_message_received(self, event: MessageReceivedEvent):
        message = event.message
        if message.__class__ in self._MESSAGE_MAP:
            await self._MESSAGE_MAP[message.__class__](message, event.connection)

    @on_message(AddUser.Response)
    async def _on_add_user(self, message: AddUser.Response, connection: PeerConnection):
        self.request_management_cycle(_RequestFlag.TRANSFER_CHANGE)

    @on_message(GetUserStatus.Response)
    async def _on_get_user_status(self, message: GetUserStatus.Response, connection: PeerConnection):
        if message.status == UserStatus.OFFLINE.value:
            self._reset_remotely_queued_flags(message.username)

        self.request_management_cycle(_RequestFlag.TRANSFER_CHANGE)

    @on_message(PeerTransferQueue.Request)
    async def _on_peer_transfer_queue(self, message: PeerTransferQueue.Request, connection: PeerConnection):
        """The peer is requesting to transfer a file to them or at least put it
        in the queue. This is usually the first message in the transfer process.

        This method will add a new transfer object to the list of transfer but
        will also check if the file actually does exist before putting it in the
        queue.
        """
        if not connection.username:
            logger.warning(
                "got PeerTransferQueue for a connection that wasn't properly initialized")
            return

        username = connection.username
        filename = message.filename
        direction = TransferDirection.UPLOAD

        if self._settings.users.is_blocked(username, BlockingFlag.UPLOADS):
            connection.queue_message(
                PeerTransferQueueFailed.Request(
                    filename=filename,
                    reason=FailReason.FILE_NOT_SHARED
                )
            )
            return

        transfer = self.find_transfer(username, filename, direction)
        fail_reason = None

        if not transfer:
            try:
                transfer = await self._add_upload(username, filename)

            except (FileNotFoundError, FileNotSharedError):
                fail_reason = FailReason.FILE_NOT_SHARED

            else:
                await transfer.state.queue()

        else:
            shared_item = await self._shares_manager.find_shared_item(
                filename,
                username
            )
            if not shared_item:
                await transfer.state.fail(FailReason.FILE_NOT_SHARED)
                fail_reason = FailReason.FILE_NOT_SHARED

            else:
                # Only put the transfer back in queue if in states:
                # * FAILED: Re-attempt at downloader getting the file
                # * COMPLETE: Re-download of a file
                # In case the state is ABORTED:
                # * Explicitly reply that the transfer has been cancelled
                # In other cases where the message is not allowed (statuses=QUEUED, INITIALIZING, UPLOADING)
                # do not respond with anything. This could otherwise abort the already
                # processing transfer
                if transfer.state.VALUE == TransferState.ABORTED:
                    fail_reason = FailReason.CANCELLED

                elif transfer.state.VALUE in (TransferState.FAILED, TransferState.COMPLETE):
                    await transfer.state.queue()

        if fail_reason:
            connection.queue_message(
                PeerTransferQueueFailed.Request(
                    filename=filename,
                    reason=fail_reason
                )
            )

    async def _on_peer_initialized(self, event: PeerInitializedEvent):
        # Only create a task for file connections that we did not try to create
        # ourselves. This (usually) means the peer is connecting to us to upload
        # a file.
        # We should first wait for the ticket to be able to link this connection
        # to a transfer
        connection = event.connection
        if connection.connection_type == PeerConnectionType.FILE:
            if event.requested:
                return

            try:
                async with atimeout(5):
                    ticket = await connection.receive_transfer_ticket()

            except (ConnectionReadError, asyncio.TimeoutError) as exc:
                # Connection should automatically be closed
                logger.warning(
                    "failed to receive transfer ticket on file connection : %s:%d",
                    connection.hostname, connection.port,
                    exc_info=exc
                )
                return

            try:
                self._file_connection_futures[ticket].set_result(connection)

            except KeyError:
                logger.warning("did not find a task waiting for file connection with ticket : %d", ticket)
                await connection.disconnect(CloseReason.REQUESTED)

            except asyncio.InvalidStateError:
                logger.warning("file connection for ticket was already fulfilled : %d", ticket)
                await connection.disconnect(CloseReason.REQUESTED)

    async def _on_session_initialized(self, event: SessionInitializedEvent):
        self.request_management_cycle(_RequestFlag.TRANSFER_CHANGE)

    @on_message(PeerTransferRequest.Request)
    async def _on_peer_transfer_request(self, message: PeerTransferRequest.Request, connection: PeerConnection):
        """The PeerTransferRequest message is sent when the peer is ready to
        transfer the file. The message contains more information about the
        transfer.

        We also handle situations here where the other peer sends this message
        without sending PeerTransferQueue first
        """
        if not connection.username:
            logger.warning(
                "got PeerTransferRequest for a connection that wasn't properly initialized")
            return

        ticket = message.ticket
        username = connection.username
        filename = message.filename
        direction = TransferDirection(message.direction)

        if self._settings.users.is_blocked(username, BlockingFlag.UPLOADS) and direction == TransferDirection.UPLOAD:
            connection.queue_message(
                PeerTransferReply.Request(
                    ticket=message.ticket,
                    allowed=False,
                    reason=FailReason.FILE_NOT_SHARED
                )
            )
            return

        transfer = self.find_transfer(username, filename, direction)

        # Make a decision based on what was requested and what we currently have
        # in our queue
        if direction == TransferDirection.UPLOAD:
            # The other peer is asking us to upload a file. Check if this is not
            # a locked file for the given user and if the item even exists
            if not transfer:
                try:
                    transfer = await self._add_upload(username, filename)

                except (FileNotFoundError, FileNotSharedError):
                    await connection.send_message(
                        PeerTransferReply.Request(
                            ticket=ticket,
                            allowed=False,
                            reason=FailReason.FILE_NOT_SHARED
                        )
                    )

                else:
                    # Send before queueing: queueing will trigger the transfer
                    # manager to re-asses the tranfers and possibly immediatly start
                    # the upload
                    connection.queue_message(
                        PeerTransferReply.Request(
                            ticket=ticket,
                            allowed=False,
                            reason=FailReason.QUEUED
                        )
                    )
                    await transfer.state.queue()

            else:
                # The peer is asking us to upload a file already in our list:
                # this always leads to a refusal of the the request as it is up
                # to us to let the downloader know when we are ready to upload
                shared_item = await self._shares_manager.find_shared_item(
                    filename,
                    username
                )
                if not shared_item:
                    await transfer.state.fail(FailReason.FILE_NOT_SHARED)
                    await connection.send_message(
                        PeerTransferReply.Request(
                            ticket=ticket,
                            allowed=False,
                            reason=FailReason.FILE_NOT_SHARED
                        )
                    )
                    return

                # If the state of transfer is PAUSED, ABORTED, COMPLETE, QUEUED:
                # refuse the request with a specific message
                # In other cases (INITIALIZING, UPLOADING, FAILED) the message
                # simply gets ignored
                # Additional notes:
                # * COMPLETE / FAILED, in reality we could re-queue here
                # * Other clients might abort the transfer when receiving this
                #   message in (at least) UPLOADING state
                fail_reason_map = {
                    TransferState.PAUSED: FailReason.CANCELLED,
                    TransferState.ABORTED: FailReason.CANCELLED,
                    TransferState.COMPLETE: FailReason.COMPLETE,
                    TransferState.QUEUED: FailReason.QUEUED
                }

                reason = fail_reason_map.get(transfer.state.VALUE, None)

                if reason:
                    await connection.send_message(
                        PeerTransferReply.Request(
                            ticket=ticket,
                            allowed=False,
                            reason=reason
                        )
                    )

        else:
            # Request to download (other peer is offering to upload a file)
            # If the transfer is not found, assume we aborted and let the
            # uploader know
            # If the transfer is ABORTED/PAUSED: let the uploader know
            # If the transfer was already COMPLETE: let the upload know
            # If the transfer is currently being processed, ignore the message
            # altogether
            # If the transfer is in FAILED state it needs to be re-queued first;
            # somehow we got this message before being able to queue the
            # transfer. Special note: because this message was received we
            # already know the transfer is remotely queued and set the flag
            # accordingly
            # Finally: if the transfer is in states QUEUED or INCOMPLETE
            # continue with the transfer as we were expecting this message

            reason = None
            if not transfer:
                # A download which we don't have in queue, assume we removed it
                reason = FailReason.CANCELLED
            else:
                current_state = transfer.state.VALUE
                if current_state in (TransferState.ABORTED, TransferState.PAUSED):
                    reason = FailReason.CANCELLED
                elif current_state == TransferState.COMPLETE:
                    reason = FailReason.COMPLETE
                elif transfer.is_processing():
                    # Needs investigation, currently don't do anything when the
                    # transfer is already being processed
                    return
                else:
                    # All good to download

                    if current_state == TransferState.FAILED:
                        await transfer.state.queue(remotely=True)

                    transfer._transfer_task = asyncio.create_task(
                        self._initialize_download(transfer, connection, message),
                        name=f'initialize-download-{task_counter()}'
                    )
                    transfer._transfer_task.add_done_callback(
                        transfer._transfer_task_complete
                    )
                    return

            await connection.send_message(
                PeerTransferReply.Request(
                    ticket=ticket,
                    allowed=False,
                    reason=reason
                )
            )

    @on_message(PeerPlaceInQueueRequest.Request)
    async def _on_peer_place_in_queue_request(
            self, message: PeerPlaceInQueueRequest.Request, connection: PeerConnection):

        if not connection.username:
            logger.warning(
                "got PeerPlaceInQueueRequest for a connection that wasn't properly initialized")
            return

        filename = message.filename
        username = connection.username
        if transfer := self.find_transfer(username, filename, TransferDirection.UPLOAD):
            place = self.get_place_in_queue(transfer)
            if place:
                connection.queue_message(
                    PeerPlaceInQueueReply.Request(filename, place))

        else:
            logger.warning(
                "PeerPlaceInQueueRequest : could not find transfer (upload) for %s from %s",
                filename,
                connection.username
            )

    @on_message(PeerPlaceInQueueReply.Request)
    async def _on_peer_place_in_queue_reply(
            self, message: PeerPlaceInQueueReply.Request, connection: PeerConnection):

        if not connection.username:
            logger.warning(
                "got PeerPlaceInQueueReply for a connection that wasn't properly initialized")
            return

        filename = message.filename
        username = connection.username
        if transfer := self.find_transfer(username, filename, TransferDirection.DOWNLOAD):
            transfer.place_in_queue = message.place

        else:
            logger.warning(
                "PeerPlaceInQueueReply : could not find transfer (download) for %s from %s",
                message.filename,
                connection.username
            )

    @on_message(PeerUploadFailed.Request)
    async def _on_peer_upload_failed(self, message: PeerUploadFailed.Request, connection: PeerConnection):
        """Called when there is a problem on their end uploading the file. This
        is actually a common message that happens for example when we close the
        connection before the upload is finished
        """
        if not connection.username:
            logger.warning(
                "got PeerUploadFailed for a connection that wasn't properly initialized")
            return

        filename = message.filename
        username = connection.username
        if transfer := self.find_transfer(username, filename, TransferDirection.DOWNLOAD):
            transfer.remotely_queued = False
            self.request_management_cycle(_RequestFlag.TRANSFER_CHANGE)

        else:
            logger.warning(
                "PeerUploadFailed : could not find transfer (download) for %s from %s",
                message.filename,
                connection.username
            )

    @on_message(PeerTransferQueueFailed.Request)
    async def _on_peer_transfer_queue_failed(
            self, message: PeerTransferQueueFailed.Request, connection: PeerConnection):

        if not connection.username:
            logger.warning(
                "got PeerTransferQueueFailed for a connection that wasn't properly initialized")
            return

        username = connection.username
        filename = message.filename
        if transfer := self.find_transfer(username, filename, TransferDirection.DOWNLOAD):
            await transfer.state.fail(reason=message.reason)

        else:
            logger.warning(
                "PeerTransferQueueFailed : could not find transfer for %s from %s",
                filename, connection.username
            )

    async def _add_upload(self, username: str, remote_path: str) -> Transfer:
        """Adds a new upload if the desired shared item exists"""

        shared_item = await self._shares_manager.get_shared_item(
            remote_path, username)

        transfer = Transfer(username, remote_path, direction=TransferDirection.UPLOAD)
        transfer.local_path = shared_item.get_absolute_path()
        transfer.filesize = await self._shares_manager.get_filesize(shared_item)

        await self.add(transfer)
        return transfer

    def _request_shares_cycle(self, event):
        self.request_management_cycle(_RequestFlag.SHARES_CHANGE)

    def _evaluate_aborted_state(self, upload: Transfer) -> tuple[bool, Optional[str]]:
        """Evaluates the aborted state of an upload to determine whether it:

        1. Should be requeued
        2. Should be aborted
        3. Abort reason has changed

        :return: a tuple with whether the transfer should change from aborted to
            queued or the other way around and the blocked reason (None if it
            needs to be requeued)
        """

        def _is_abort_requested(transfer: Transfer) -> bool:
            return transfer.abort_reason == AbortReason.REQUESTED

        def _is_blocked(transfer: Transfer) -> bool:
            return self._settings.users.is_blocked(transfer.username, BlockingFlag.UPLOADS)

        def _is_not_shared(transfer: Transfer) -> bool:
            return not bool(self._shares_manager.find_shared_item_cache(
                transfer.remote_path, transfer.username
            ))

        conditions = (
            (_is_abort_requested, AbortReason.REQUESTED),
            (_is_blocked, AbortReason.BLOCKED),
            (_is_not_shared, AbortReason.FILE_NOT_SHARED)
        )
        abort_reason = None
        for condition, reason in conditions:
            if condition(upload):
                abort_reason = reason
                break

        aborted = upload.state.VALUE == TransferState.ABORTED
        should_change = aborted != bool(abort_reason)

        return (should_change, abort_reason)



from abc import ABC
import asyncio


class BaseManager(ABC):
    """Base class for manager instances. This class provides hooks that will
    get called during starting and stopping of the client
    """

    async def load_data(self):
        """Load the data for the manager. This method gets called during client
        client start-up before any connection is made
        """

    async def store_data(self):
        """Stores data from the manager. This gets called during the client
        shutdown
        """

    async def start(self):
        """Optionally performs a start action on the manager. Gets called after
        loading the data but before connecting
        """

    async def stop(self) -> list[asyncio.Task]:
        """Cancel all running tasks. The implementation of this method should
        simply cancel the task and return, the code calling this method should
        be responsible for awaiting the cancellation

        :return: list of all cancelled tasks
        """
        return []



import asyncio
from collections.abc import Callable, Coroutine
import logging
from typing import Any, Optional, Union


logger = logging.getLogger(__name__)


Interval = Union[float, Callable[[], float]]
TaskCoroutine = Union[
    Callable[[], Coroutine[None, None, Optional[float]]],
    Callable[[Any], Coroutine[None, None, Optional[float]]]
]


class BackgroundTask:
    """Wrapper class for a background task

    :ivar interval: interval at which the task should run
    :ivar task_coro: coroutine function that should be run at the provided
        interval
    :ivar preempt_wait: whether to perform an initial wait before running the
        coroutine for the first time
    :ivar name: optional task name
    :ivar context: optional context to pass to the coroutine
    """

    def __init__(
            self, interval: Interval, task_coro: TaskCoroutine,
            preempt_wait: bool = False, name: str = 'background-task',
            context: Optional[Any] = None):

        self.interval: Interval = interval
        self.preempt_wait: bool = preempt_wait
        self.task_coro: TaskCoroutine = task_coro
        self.name: str = name
        self.context: Optional[Any] = context
        self._task: Optional[asyncio.Task] = None

    def resolve_interval(self) -> float:
        if callable(self.interval):
            return self.interval()

        return self.interval

    def start(self):
        if not self._task:
            logger.info("starting task : %s", self.name)
            self._task = asyncio.create_task(self.runner(), name=self.name)

    def cancel(self) -> Optional[asyncio.Task]:
        if not self._task:
            return None

        logger.info("cancelling task : %s", self.name)
        task = self._task
        self._task = None
        task.cancel()
        return task

    def is_running(self) -> bool:
        return bool(self._task)

    async def runner(self):
        if self.preempt_wait:
            await asyncio.sleep(self.resolve_interval())

        while True:
            coro = self.task_coro(self.context) if self.context else self.task_coro()  # type: ignore[call-arg]
            job_ival = await coro

            next_ival = self.resolve_interval() if job_ival is None else job_ival

            await asyncio.sleep(next_ival)


class Timer:
    """Re-schedulable timer which runs a callback after completion"""

    def __init__(self, timeout: float, callback: TaskCoroutine):
        self.timeout: float = timeout
        self.callback: TaskCoroutine = callback
        self._task: Optional[asyncio.Task] = None

    def start(self):
        self._task = asyncio.create_task(self.runner())
        self._task.add_done_callback(self._unset_task)

    def cancel(self) -> Optional[asyncio.Task]:
        if self._task is None:
            return None

        task = self._task
        self._task.cancel()
        self._task = None
        return task

    async def runner(self):
        await asyncio.sleep(self.timeout)
        await self.callback()  # type: ignore[call-arg]

    def reschedule(self, timeout: Optional[float] = None):
        self.timeout = self.timeout if timeout is None else timeout

        self.cancel()
        self.start()

    def _unset_task(self, task: asyncio.Future):
        self._task = None



import logging

from .base_manager import BaseManager
from .constants import SERVER_PING_INTERVAL
from .events import ConnectionStateChangedEvent, EventBus
from .protocol.messages import Ping
from .network.connection import ConnectionState, ServerConnection
from .network.network import Network
from .settings import Settings
from .tasks import BackgroundTask


logger = logging.getLogger(__name__)


class ServerManager(BaseManager):
    """Class handling server state changes"""

    def __init__(self, settings: Settings, event_bus: EventBus, network: Network):
        self._settings: Settings = settings
        self._event_bus: EventBus = event_bus
        self._network: Network = network

        self._ping_task: BackgroundTask = BackgroundTask(
            interval=SERVER_PING_INTERVAL,
            task_coro=self._ping_job,
            preempt_wait=True,
            name='server-ping-task'
        )

        self.register_listeners()

    @property
    def connection_state(self) -> ConnectionState:
        return self._network.server_connection.state

    def register_listeners(self):
        self._event_bus.register(
            ConnectionStateChangedEvent, self._on_state_changed)

    async def send_ping(self):
        """Send ping to the server"""
        await self._network.send_server_messages(Ping.Request())

    # Job methods
    async def _ping_job(self):
        try:
            await self.send_ping()
        except Exception as exc:
            logger.warning("failed to ping server. exception=%r", exc)

    # Listeners

    async def _on_state_changed(self, event: ConnectionStateChangedEvent):
        if not isinstance(event.connection, ServerConnection):
            return

        if event.state == ConnectionState.CONNECTED:
            self._ping_task.start()

        elif event.state == ConnectionState.CLOSING:
            self._ping_task.cancel()
